"""MCP9808 temp sensor."""

from adafruit_mcp9808 import MCP9808

from boneio.const import MCP_TEMP_9808

from . import TempSensor


class MCP9808Sensor(TempSensor):
    """Represent MCP9808 sensor in BoneIO."""

    SensorClass = MCP9808
    DefaultName = MCP_TEMP_9808



"""Runner code for boneIO. Based on HA runner."""
import asyncio
import logging
import os

from boneio.const import (
    ADC,
    COVER,
    DALLAS,
    DS2482,
    ENABLED,
    HA_DISCOVERY,
    HOST,
    INPUT,
    LM75,
    MCP23017,
    MCP_TEMP_9808,
    MODBUS,
    MQTT,
    OLED,
    ONEWIRE,
    OUTPUT,
    PASSWORD,
    PORT,
    SENSOR,
    TOPIC_PREFIX,
    USERNAME,
)
from boneio.helper import StateManager
from boneio.helper.config import ConfigHelper
from boneio.manager import Manager
from boneio.mqtt_client import MQTTClient

_LOGGER = logging.getLogger(__name__)


async def async_run(
    config: dict,
    config_file: str,
    mqttusername: str = "",
    mqttpassword: str = "",
):
    """Run BoneIO."""

    _config_helper = ConfigHelper(
        topic_prefix=config[MQTT].pop(TOPIC_PREFIX),
        ha_discovery=config[MQTT][HA_DISCOVERY].pop(ENABLED),
        ha_discovery_prefix=config[MQTT][HA_DISCOVERY].pop(TOPIC_PREFIX),
    )

    client = MQTTClient(
        host=config[MQTT][HOST],
        username=config[MQTT].get(USERNAME, mqttusername),
        password=config[MQTT].get(PASSWORD, mqttpassword),
        port=config[MQTT].get(PORT, 1883),
        config_helper=_config_helper,
    )

    manager = Manager(
        send_message=client.send_message,
        relay_pins=config.get(OUTPUT, []),
        input_pins=config.get(INPUT, []),
        config_file_path=config_file,
        state_manager=StateManager(
            state_file=f"{os.path.split(config_file)[0]}state.json"
        ),
        config_helper=_config_helper,
        sensors={
            LM75: config.get(LM75, []),
            MCP_TEMP_9808: config.get(MCP_TEMP_9808, []),
            MODBUS: config.get("modbus_sensors"),
            ONEWIRE: config.get(SENSOR, []),
        },
        mcp23017=config.get(MCP23017, []),
        ds2482=config.get(DS2482, []),
        dallas=config.get(DALLAS),
        modbus=config.get(MODBUS),
        oled=config.get(OLED),
        adc_list=config.get(ADC, []),
        covers=config.get(COVER, []),
    )
    tasks = set()
    tasks.update(manager.get_tasks())
    _LOGGER.info("Connecting to MQTT.")
    tasks.add(client.start_client(manager))
    return await asyncio.gather(*tasks)



from __future__ import annotations

import logging
import time
from collections import namedtuple
from typing import TYPE_CHECKING, Any, Callable, Dict, Union

from adafruit_mcp230xx.mcp23017 import MCP23017

from boneio.const import (
    ACTIONS,
    ADDRESS,
    BINARY_SENSOR,
    COVER,
    DEVICE_CLASS,
    FILTERS,
    GPIO,
    ID,
    INIT_SLEEP,
    INPUT,
    INPUT_SENSOR,
    KIND,
    LM75,
    MCP,
    MCP_ID,
    MCP_TEMP_9808,
    MODEL,
    NONE,
    OUTPUT_TYPE,
    PIN,
    RELAY,
    RESTORE_STATE,
    SENSOR,
    SHOW_HA,
    UPDATE_INTERVAL,
    DallasBusTypes,
)
from boneio.cover import Cover
from boneio.helper import (
    GPIOInputException,
    I2CError,
    StateManager,
    ha_adc_sensor_availabilty_message,
    ha_binary_sensor_availabilty_message,
    ha_input_availabilty_message,
    ha_sensor_temp_availabilty_message,
)
from boneio.helper.onewire import (
    DS2482,
    DS2482_ADDRESS,
    OneWireBus,
    AsyncBoneIOW1ThermSensor,
    OneWireAddress,
)
from boneio.helper.ha_discovery import ha_cover_availabilty_message
from boneio.helper.timeperiod import TimePeriod
from boneio.input.gpio import GpioInputButton
from boneio.sensor import DallasSensorDS2482
from boneio.sensor.temp.dallas import DallasSensorW1

# Typing imports that create a circular dependency
if TYPE_CHECKING:
    from ..manager import Manager

from busio import I2C

from boneio.relay import GpioRelay, MCPRelay
from boneio.sensor import GpioADCSensor, initialize_adc
from boneio.sensor.gpio import GpioInputSensor

_LOGGER = logging.getLogger(__name__)


def create_adc(manager: Manager, topic_prefix: str, adc_list: list = []):
    """Create ADC sensor."""

    initialize_adc()

    # TODO: find what exception can ADC gpio throw.
    for gpio in adc_list:
        name = gpio.get(ID)
        id = name.replace(" ", "")
        pin = gpio[PIN]
        try:
            GpioADCSensor(
                id=id,
                pin=pin,
                name=name,
                manager=manager,
                send_message=manager.send_message,
                topic_prefix=topic_prefix,
                update_interval=gpio.get(UPDATE_INTERVAL, TimePeriod(seconds=60)),
            )
            if gpio.get(SHOW_HA, True):
                manager.send_ha_autodiscovery(
                    id=id,
                    name=name,
                    ha_type=SENSOR,
                    availability_msg_func=ha_adc_sensor_availabilty_message,
                )
        except I2CError as err:
            _LOGGER.error("Can't configure ADC sensor %s. %s", id, err)
            pass


def create_temp_sensor(
    manager: Manager,
    topic_prefix: str,
    sensor_type: str,
    i2cbusio: I2C,
    config: dict = {},
):
    """Create LM sensor in manager."""
    if sensor_type == LM75:
        from boneio.sensor import LM75Sensor as TempSensor
    elif sensor_type == MCP_TEMP_9808:
        from boneio.sensor import MCP9808Sensor as TempSensor
    else:
        return
    name = config.get(ID)
    id = name.replace(" ", "")
    try:
        temp_sensor = TempSensor(
            id=id,
            name=name,
            i2c=i2cbusio,
            address=config[ADDRESS],
            manager=manager,
            send_message=manager.send_message,
            topic_prefix=topic_prefix,
            update_interval=config.get(UPDATE_INTERVAL, TimePeriod(seconds=60)),
            filters=config.get(FILTERS, []),
        )
        manager.send_ha_autodiscovery(
            id=id,
            name=name,
            ha_type=SENSOR,
            availability_msg_func=ha_sensor_temp_availabilty_message,
        )
        return temp_sensor
    except I2CError as err:
        _LOGGER.error("Can't configure Temp sensor. %s", err)
        pass


def create_mcp23017(
    manager: Manager,
    mcp23017: list,
    i2cbusio: I2C,
) -> dict:
    """Create MCP23017."""
    grouped_outputs = {}
    for mcp in mcp23017:
        id = mcp[ID] or mcp[ADDRESS]
        try:
            manager._mcp[id] = MCP23017(i2c=i2cbusio, address=mcp[ADDRESS], reset=False)
            sleep_time = mcp.get(INIT_SLEEP, TimePeriod(seconds=0))
            _LOGGER.debug(
                f"Sleeping for {sleep_time.total_seconds}s while MCP {id} is initializing."
            )
            time.sleep(sleep_time.total_seconds)
            grouped_outputs[id] = {}
        except TimeoutError as err:
            _LOGGER.error("Can't connect to MCP %s. %s", id, err)
            pass
    return grouped_outputs


def create_modbus_sensors(manager: Manager, sensors, **kwargs) -> None:
    """Create Modbus sensor for each device."""
    from boneio.sensor.modbus import ModbusSensor

    for sensor in sensors:
        name = sensor.get(ID)
        id = name.replace(" ", "")
        try:
            ModbusSensor(
                address=sensor[ADDRESS],
                id=id,
                name=name,
                manager=manager,
                model=sensor[MODEL],
                send_message=manager.send_message,
                update_interval=sensor.get(UPDATE_INTERVAL, TimePeriod(seconds=60)),
                **kwargs,
            )
        except FileNotFoundError as err:
            _LOGGER.error(
                "Can't configure Modbus sensor %s. %s. No such model in database.",
                id,
                err,
            )
            pass


OutputEntry = namedtuple("OutputEntry", "OutputClass output_kind output_id")


def output_chooser(output_kind: str, mcp_id: str | None):
    """Get named tuple based on input."""
    if output_kind == MCP:
        return OutputEntry(MCPRelay, MCP, mcp_id)
    else:
        return OutputEntry(GpioRelay, GPIO, GPIO)


def configure_relay(
    manager: Manager,
    state_manager: StateManager,
    topic_prefix: str,
    relay_id: str,
    relay_callback: Callable,
    config: dict,
) -> Any:
    """Configure kind of relay. Most common MCP."""
    restore_state = config.pop(RESTORE_STATE, False)
    output_type = config.pop(OUTPUT_TYPE)
    restored_state = (
        state_manager.get(attr_type=RELAY, attr=relay_id, default_value=False)
        if restore_state
        else False
    )
    if output_type == NONE and state_manager.get(attr_type=RELAY, attr=relay_id):
        state_manager.del_attribute(attr_type=RELAY, attribute=relay_id)
        restored_state = False

    output = output_chooser(
        output_kind=config.pop(KIND), mcp_id=config.pop(MCP_ID, None)
    )

    if getattr(output, "output_kind") == MCP:
        mcp = manager.mcp.get(getattr(output, "output_id"))
        if not mcp:
            _LOGGER.error("No such MCP configured!")
            return None
        kwargs = {
            "pin": int(config.pop(PIN)),
            "mcp": mcp,
            "mcp_id": getattr(output, "output_id"),
            "output_type": output_type,
        }
    elif getattr(output, "output_kind") == GPIO:
        if GPIO not in manager.grouped_outputs:
            manager.grouped_outputs[GPIO] = {}
        kwargs = {
            "pin": config.pop(PIN),
        }
    else:
        return
    relay = getattr(output, "OutputClass")(
        send_message=manager.send_message,
        topic_prefix=topic_prefix,
        id=config.pop(ID),
        restored_state=restored_state,
        **config,
        **kwargs,
        callback=lambda: relay_callback(
            relay_type=getattr(output, "output_id"),
            relay_id=relay_id,
            restore_state=False if output_type == NONE else restore_state,
        ),
    )
    manager.grouped_outputs[getattr(output, "output_id")][relay_id] = relay
    return relay


InputEntry = namedtuple(
    "InputEntry", "InputClass input_type ha_type availability_msg_f"
)


def input_chooser(input_type: str):
    """Get named tuple based on input."""
    if input_type == SENSOR:
        return InputEntry(
            GpioInputSensor,
            INPUT_SENSOR,
            BINARY_SENSOR,
            ha_binary_sensor_availabilty_message,
        )
    else:
        return InputEntry(GpioInputButton, INPUT, SENSOR, ha_input_availabilty_message)


def configure_input(
    gpio: dict,
    pin: str,
    press_callback: Callable,
    send_ha_autodiscovery: Callable,
) -> str:
    """Configure input sensor or button."""
    try:
        input = input_chooser(input_type=gpio.get(KIND))
        getattr(input, "InputClass")(
            pin=pin,
            press_callback=lambda x, i: press_callback(
                x=x,
                inpin=i,
                actions=gpio.get(ACTIONS, {}).get(x, []),
                input_type=getattr(input, "input_type"),
            ),
            rest_pin=gpio,
        )
        if gpio.get(SHOW_HA, True):
            send_ha_autodiscovery(
                id=pin,
                name=gpio.get(ID, pin),
                ha_type=getattr(input, "ha_type"),
                availability_msg_func=getattr(input, "availability_msg_f"),
            )
        return pin
    except GPIOInputException as err:
        _LOGGER.error("This PIN %s can't be configured. %s", pin, err)
        pass


def configure_cover(
    manager: Manager,
    cover_id: str,
    state_manager: StateManager,
    send_ha_autodiscovery: Callable,
    config: dict,
    **kwargs,
) -> Cover:
    restored_state = state_manager.get(
        attr_type=COVER, attr=cover_id, default_value=100
    )

    def state_save(position: int):
        if config[RESTORE_STATE]:
            state_manager.save_attribute(
                attr_type=COVER,
                attribute=cover_id,
                value=position,
            )

    cover = Cover(
        id=cover_id,
        state_save=state_save,
        send_message=manager.send_message,
        restored_state=restored_state,
        **kwargs,
    )
    if config.get(SHOW_HA, True):
        send_ha_autodiscovery(
            id=cover.id,
            name=cover.name,
            ha_type=COVER,
            device_class=config.get(DEVICE_CLASS),
            availability_msg_func=ha_cover_availabilty_message,
        )
    return cover


def configure_ds2482(i2cbusio: I2C, address: str = DS2482_ADDRESS) -> OneWireBus:
    ds2482 = DS2482(i2c=i2cbusio, address=address)
    ow_bus = OneWireBus(ds2482=ds2482)
    return ow_bus


def configure_dallas() -> AsyncBoneIOW1ThermSensor:
    return AsyncBoneIOW1ThermSensor


def find_onewire_devices(
    ow_bus: Union[OneWireBus, AsyncBoneIOW1ThermSensor],
    bus_id: str,
    bus_type: DallasBusTypes,
) -> Dict[OneWireAddress]:
    out = {}
    try:
        devices = ow_bus.scan()
        for device in devices:
            _addr: int = device.int_address
            _LOGGER.debug("Found device on bus %s with address %s", bus_id, hex(_addr))
            out[_addr] = device
    except RuntimeError as err:
        _LOGGER.error("Problem with scanning %s bus. %s", bus_type, err)
    return out


def create_dallas_sensor(
    manager: Manager,
    address: OneWireAddress,
    config: dict,
    **kwargs,
) -> Union[DallasSensorDS2482, DallasSensorW1]:
    name = config.get(ID) or hex(address)
    id = name.replace(" ", "")
    bus: OneWireBus = kwargs.get("bus")
    cls = DallasSensorDS2482 if bus else DallasSensorW1
    sensor = cls(
        manager=manager,
        address=address,
        id=id,
        name=name,
        update_interval=config.get(UPDATE_INTERVAL, TimePeriod(seconds=60)),
        send_message=manager.send_message,
        filters=config.get(FILTERS, []),
        **kwargs,
    )
    if config.get(SHOW_HA, True):
        manager.send_ha_autodiscovery(
            id=sensor.id,
            name=sensor.name,
            ha_type=SENSOR,
            availability_msg_func=ha_sensor_temp_availabilty_message,
        )
    return sensor



"""
Provide an MQTT client for providing BoneIO MQTT broker.
Code based on cgarwood/python-openzwave-mqtt.
"""
import asyncio
import json
import logging
import uuid
from contextlib import AsyncExitStack
from typing import Any, Callable, Optional, Set, Tuple, Union

import paho.mqtt.client as mqtt
from asyncio_mqtt import Client as AsyncioClient
from asyncio_mqtt import MqttError, Will
from paho.mqtt.properties import Properties
from paho.mqtt.subscribeoptions import SubscribeOptions

from boneio.const import ONLINE, PAHO, STATE
from boneio.helper import UniqueQueue
from boneio.helper.config import ConfigHelper

_LOGGER = logging.getLogger(__name__)


class MQTTClient:
    """Represent an MQTT client."""

    def __init__(
        self,
        host: str,
        config_helper: ConfigHelper,
        port: int = 1883,
        **client_options: Any,
    ) -> None:
        """Set up client."""
        self.host = host
        self.port = port
        self._config_helper = config_helper
        client_options["client_id"] = mqtt.base62(uuid.uuid4().int, padding=22)
        client_options["logger"] = logging.getLogger(PAHO)
        client_options["clean_session"] = True
        self.client_options = client_options
        self.asyncio_client: AsyncioClient = None
        self.create_client()
        self.reconnect_interval = 1
        self._connection_established = False
        self.publish_queue: UniqueQueue = UniqueQueue()

    def create_client(self) -> None:
        """Create the asyncio client."""
        _LOGGER.debug("Creating client %s:%s", self.host, self.port)
        self.asyncio_client = AsyncioClient(
            self.host,
            self.port,
            will=Will(
                topic=f"{self._config_helper.topic_prefix}/{STATE}",
                payload=ONLINE,
                qos=0,
                retain=False,
            ),
            **self.client_options,
        )

    async def publish(  # pylint:disable=too-many-arguments
        self,
        topic: str,
        payload: Optional[str] = None,
        retain: bool = False,
        qos: int = 0,
        properties: Optional[Properties] = None,
        timeout: float = 10,
    ) -> None:
        """Publish to topic.

        Can raise asyncio_mqtt.MqttError.
        """
        params: dict = {"qos": qos, "retain": retain, "timeout": timeout}
        if payload:
            params["payload"] = payload
        if properties:
            params["properties"] = properties

        _LOGGER.debug("Sending message topic: %s, payload: %s", topic, payload)
        await self.asyncio_client.publish(topic, **params)

    async def subscribe(  # pylint:disable=too-many-arguments
        self,
        topics: Tuple[str, str],
        qos: int = 0,
        options: Optional[SubscribeOptions] = None,
        properties: Optional[Properties] = None,
        timeout: float = 10.0,
    ) -> None:
        """Subscribe to topic.

        Can raise asyncio_mqtt.MqttError.
        """
        args = []
        for topic in topics:
            args.append((topic, qos))
        params: dict = {"qos": qos}
        if options:
            params["options"] = options
        if properties:
            params["properties"] = properties

        # e.g. subscribe([("my/topic", SubscribeOptions(qos=0), ("another/topic", SubscribeOptions(qos=2)])
        await self.asyncio_client.subscribe(topic=args, **params, timeout=timeout)

    async def unsubscribe(
        self, topic: str, properties: Optional[Properties] = None, timeout: float = 10.0
    ) -> None:
        """Unsubscribe from topic.

        Can raise asyncio_mqtt.MqttError.
        """
        params: dict = {"timeout": timeout}
        if properties:
            params["properties"] = properties

        await self.asyncio_client.unsubscribe(topic, **params)

    def send_message(
        self, topic: str, payload: Union[str, dict], retain: bool = False
    ) -> None:
        """Send a message from the manager options."""
        to_publish = (
            topic,
            json.dumps(payload) if type(payload) == dict else payload,
            retain,
        )
        self.publish_queue.put_nowait(to_publish)

    async def _handle_publish(self) -> None:
        """Publish messages as they are put on the queue."""
        while True:
            to_publish: tuple = await self.publish_queue.get()
            await self.publish(*to_publish)
            self.publish_queue.task_done()

    async def start_client(self, manager: any) -> None:
        """Start the client with the manager."""
        # Reconnect automatically until the client is stopped.
        while True:
            try:
                await self._subscribe_manager(manager)
            except MqttError as err:
                self.reconnect_interval = min(self.reconnect_interval * 2, 900)
                _LOGGER.error(
                    "MQTT error: %s. Reconnecting in %s seconds",
                    err,
                    self.reconnect_interval,
                )
                self._connection_established = False
                await asyncio.sleep(self.reconnect_interval)
                self.create_client()  # reset connect/reconnect futures

    async def _subscribe_manager(self, manager: any) -> None:
        """Connect and subscribe to manager topics + host stats."""
        async with AsyncExitStack() as stack:
            tasks: Set[asyncio.Task] = set()

            # Connect to the MQTT broker.
            await stack.enter_async_context(self.asyncio_client)
            # Reset the reconnect interval after successful connection.
            self.reconnect_interval = 1

            publish_task = asyncio.create_task(self._handle_publish())
            tasks.add(publish_task)

            # Messages that doesn't match a filter will get logged and handled here.
            messages = await stack.enter_async_context(
                self.asyncio_client.unfiltered_messages()
            )

            messages_task = asyncio.create_task(
                handle_messages(messages, manager.receive_message)
            )
            if not self._connection_established:
                self._connection_established = True
                reconnect_task = asyncio.create_task(manager.reconnect_callback())
                tasks.add(reconnect_task)
            tasks.add(messages_task)

            await self.subscribe(
                topics=(self._config_helper.subscribe_topic, "homeassistant/status")
            )

            # Wait for everything to complete (or fail due to, e.g., network errors).
            await asyncio.gather(*tasks)


async def handle_messages(messages: Any, callback: Callable[[str, str], None]) -> None:
    """Handle messages with callback."""
    async for message in messages:
        payload = message.payload.decode()
        _LOGGER.debug("Received message topic: %s, payload: %s", message.topic, payload)
        await callback(message.topic, payload)



import asyncio
import datetime as dt
import logging
import signal
import time
from datetime import datetime
from typing import Callable, Optional

from boneio.helper.util import callback

_LOGGER = logging.getLogger(__name__)
UTC = dt.timezone.utc
EVENT_TIME_CHANGED = "event_time_changed"
CALLBACK_TYPE = Callable[[], None]


def utcnow() -> dt.datetime:
    return dt.datetime.now(UTC)


time_tracker_utcnow = utcnow


def _async_create_timer(loop: asyncio.AbstractEventLoop, event_callback) -> None:
    """Create a timer that will start on BoneIO start."""
    handle = None

    def schedule_tick(now: dt.datetime) -> None:
        """Schedule a timer tick when the next second rolls around."""
        nonlocal handle
        slp_seconds = 1 - (now.microsecond / 10**6)
        handle = loop.call_later(slp_seconds, fire_time_event)

    def fire_time_event() -> None:
        """Fire next time event."""
        now = utcnow()
        event_callback(now)
        schedule_tick(now)

    def stop_timer() -> None:
        """Stop the timer."""
        if handle is not None:
            handle.cancel()

    schedule_tick(utcnow())
    return stop_timer


class GracefulExit(SystemExit):
    """Graceful exit."""

    def __init__(self, msg=None, code=None):
        super(GracefulExit, self).__init__(msg)
        self.code = code


class ListenerJob:
    """Listener to represent jobs during runtime."""

    def __init__(self, target) -> None:
        """Initialize listener."""
        self.target = target
        self._handle = None

    def add_handle(self, handle):
        """Add handle to listener."""
        self._handle = handle

    @property
    def handle(self):
        """Return handle."""
        return self._handle


class EventBus:
    """Simple event bus which ticks every second."""

    def __init__(self, loop: asyncio.AbstractEventLoop) -> None:
        """Initialize handler"""
        self._loop = loop or asyncio.get_event_loop()
        self._listeners = {}
        self._sigterm_listeners = []
        self._haonline_listeners = []
        self._timer_handle = _async_create_timer(self._loop, self._run_second_event)
        for signame in {"SIGINT", "SIGTERM"}:
            self._loop.add_signal_handler(
                getattr(signal, signame),
                self.ask_exit,
            )

    def _run_second_event(self, time):
        """Run event every second."""
        for key, listener in self._listeners.items():
            if listener.target:
                self._listeners[key].add_handle(
                    self._loop.call_soon(listener.target, time)
                )

    def ask_exit(self):
        """Function to call on exit. Should invoke all sigterm listeners."""
        _LOGGER.debug("Exiting process started.")
        self._listeners = {}
        for target in self._sigterm_listeners:
            target()
        self._timer_handle()
        _LOGGER.info("Shutdown gracefully.")
        raise GracefulExit(code=0)

    def add_listener(self, name, target):
        """Add listener on every second job."""
        self._listeners[name] = ListenerJob(target=target)
        return self._listeners[name]

    def add_sigterm_listener(self, target):
        """Add sigterm listener."""
        self._sigterm_listeners.append(target)

    def add_haonline_listener(self, target):
        """Add HA Online listener."""
        self._haonline_listeners.append(target)

    def signal_ha_online(self):
        """Call events if HA goes online."""
        for target in self._haonline_listeners:
            target()

    def remove_listener(self, name):
        """Remove regular listener."""
        if name in self._listeners:
            del self._listeners[name]


def as_utc(dattim: dt.datetime) -> dt.datetime:
    """Return a datetime as UTC time.

    Assumes datetime without tzinfo to be in the DEFAULT_TIME_ZONE.
    """
    if dattim.tzinfo == UTC:
        return dattim
    return dattim.astimezone(UTC)


@callback
def async_track_point_in_time(
    loop: asyncio.AbstractEventLoop,
    action,
    point_in_time: datetime,
) -> CALLBACK_TYPE:
    """Add a listener that fires once after a specific point in UTC time."""
    # Ensure point_in_time is UTC
    utc_point_in_time = as_utc(point_in_time)

    # Since this is called once, we accept a so we can avoid
    # having to figure out how to call the action every time its called.
    cancel_callback: Optional[asyncio.TimerHandle] = None

    @callback
    def run_action(job) -> None:
        """Call the action."""
        nonlocal cancel_callback

        now = time_tracker_utcnow()

        # Depending on the available clock support (including timer hardware
        # and the OS kernel) it can happen that we fire a little bit too early
        # as measured by utcnow(). That is bad when callbacks have assumptions
        # about the current time. Thus, we rearm the timer for the remaining
        # time.
        delta = (utc_point_in_time - now).total_seconds()
        if delta > 0:
            _LOGGER.debug("Called %f seconds too early, rearming", delta)

            cancel_callback = loop.call_later(delta, run_action, job)
            return

        loop.call_soon(job, utc_point_in_time)

    delta = utc_point_in_time.timestamp() - time.time()
    cancel_callback = loop.call_later(delta, run_action, action)

    @callback
    def unsub_point_in_time_listener() -> None:
        """Cancel the call_later."""
        assert cancel_callback is not None
        cancel_callback.cancel()

    return unsub_point_in_time_listener



"""Custom W1ThermSensor class to support hex addressing of W1 sensors.
"""
from typing import List
from w1thermsensor import AsyncW1ThermSensor, Sensor
from boneio.helper.onewire import OneWireAddress, reverse_dallas_id


def crc82(data: bytearray) -> int:
    """
    Perform the 1-Wire CRC check on the provided data. Function from Circuit Python Adafruit.

    :param bytearray data: 8 byte array representing 64 bit ROM code
    """
    crc = 0

    for byte in data:
        crc ^= byte
        for _ in range(8):
            if crc & 0x01:
                crc = (crc >> 1) ^ 0x8C
            else:
                crc >>= 1
            crc &= 0xFF
    return crc


class AsyncBoneIOW1ThermSensor(AsyncW1ThermSensor):
    def __init__(self, *args, **kwargs):  # pragma: no cover
        """Custom init function to work with same addressing type as esphome."""
        self._ds18b20_str_id = hex(Sensor.DS18B20)[2:]
        try:
            sensor_id = kwargs.pop("sensor_id")
            if type(sensor_id) == OneWireAddress:
                sensor_id = sensor_id.hw_id
            super().__init__(*args, **kwargs, sensor_id=sensor_id)
        except KeyError:
            super().__init__(*args, **kwargs)
        _crc = crc82(
            bytes.fromhex(f"{self._ds18b20_str_id}{reverse_dallas_id(self.id)}")
        )
        self._hex_id = f"{hex(_crc)}{self.id}{self._ds18b20_str_id}".lower()

    @classmethod
    def scan(cls) -> List["OneWireAddress"]:
        """Return only DS18B20 sensors. Add more sensors in the future."""

        def is_sensor(dir_name):
            return dir_name.startswith(hex(Sensor.DS18B20)[2:])

        def get_hex(name):
            _hex_id = bytes.fromhex(f"{name[:2]}{reverse_dallas_id(name[3:])}")
            return _hex_id + bytes([crc82(_hex_id)])

        return [
            OneWireAddress(get_hex(name=s.name))
            for s in cls.BASE_DIRECTORY.iterdir()
            if is_sensor(s.name)
        ]

    @property
    def rom(self) -> bytes:
        """Get rom id."""
        return bytearray.fromhex(self.hex_id)

    @property
    def hex_id(self) -> str:
        """Get hex representation of sensor."""
        if not self._hex_id:
            _crc = crc82(
                bytes.fromhex(f"{self._ds18b20_str_id}{reverse_dallas_id(self.id)}")
            )
            self._hex_id = f"{hex(_crc)}{self.id}{self._ds18b20_str_id}".lower()
        return self._hex_id



import asyncio
import logging
from itertools import cycle
from typing import List

from luma.core.error import DeviceNotFoundError
from luma.core.interface.serial import i2c
from luma.core.render import canvas
from luma.oled.device import sh1106
from PIL import ImageDraw

from boneio.const import CPU, DISK, MEMORY, NETWORK, OLED_PIN, SWAP, UPTIME, WHITE
from boneio.helper import (
    HostData,
    I2CError,
    TimePeriod,
    edge_detect,
    make_font,
    setup_input,
)
from boneio.helper.events import async_track_point_in_time, utcnow

_LOGGER = logging.getLogger(__name__)


fontBig = make_font("DejaVuSans.ttf", 12)
fontSmall = make_font("DejaVuSans.ttf", 9)
fontExtraSmall = make_font("DejaVuSans.ttf", 7)
danube = make_font("danube__.ttf", 15, local=True)

# screen_order = [UPTIME, NETWORK, CPU, DISK, MEMORY, SWAP]

STANDARD_ROWS = [17, 32, 47]
UPTIME_ROWS = list(range(22, 60, 10))
OUTPUT_ROWS = list(range(14, 60, 6))
OUTPUT_COLS = range(0, 113, 56)


class Oled:
    """Oled display class."""

    def __init__(
        self,
        host_data: HostData,
        output_groups: List[str],
        sleep_timeout: TimePeriod,
        screen_order: List[str],
    ) -> None:
        """Initialize OLED screen."""
        self._loop = asyncio.get_running_loop()
        self._output_groups = None
        try:
            _ind_screen = screen_order.index("outputs")
            if not output_groups:
                _LOGGER.debug("No outputs configured. Omitting in screen.")
                screen_order.pop(_ind_screen)
            screen_order[_ind_screen:_ind_screen] = output_groups
            self._output_groups = output_groups
        except ValueError:
            pass
        self._screen_order = cycle(screen_order)
        self._current_screen = next(self._screen_order)
        self._host_data = host_data
        self._sleep = False
        self._sleep_handle = None
        self._sleep_timeout = sleep_timeout
        setup_input(pin=OLED_PIN, pull_mode="gpio_pu")
        edge_detect(pin=OLED_PIN, callback=self._handle_press, bounce=240)
        try:
            serial = i2c(port=2, address=0x3C)
            self._device = sh1106(serial)
        except DeviceNotFoundError as err:
            raise I2CError(err)
        _LOGGER.debug("Configuring OLED screen.")

    def _draw_standard(self, data: dict, draw: ImageDraw) -> None:
        """Draw standard information about host screen."""
        draw.text((1, 1), self._current_screen, font=fontBig, fill=WHITE)
        i = 0
        for k in data:
            draw.text(
                (3, STANDARD_ROWS[i]),
                f"{k} {data[k]}",
                font=fontSmall,
                fill=WHITE,
            )
            i += 1

    def _sleeptime(self):
        with canvas(self._device) as draw:
            draw.rectangle(self._device.bounding_box, outline="black", fill="black")
        self._sleep = True

    def _draw_uptime(self, data: dict, draw: ImageDraw) -> None:
        """Draw uptime screen with boneIO logo."""
        draw.text((3, 3), "bone", font=danube, fill=WHITE)
        draw.text((53, 3), "iO", font=danube, fill=WHITE)
        i = 0
        for k in data:
            draw.text(
                (3, UPTIME_ROWS[i]),
                f"{k}: {data[k]}",
                font=fontSmall,
                fill=WHITE,
            )
            i += 1

    def _draw_output(self, data: dict, draw: ImageDraw) -> None:
        "Draw outputs of GPIO/MCP relays."
        cols = cycle(OUTPUT_COLS)
        draw.text((1, 1), f"Relay {self._current_screen}", font=fontSmall, fill=WHITE)
        i = 0
        j = next(cols)
        for k in data:
            if len(OUTPUT_ROWS) == i:
                j = next(cols)
                i = 0
            draw.text(
                (j, OUTPUT_ROWS[i]), f"{k} {data[k]}", font=fontExtraSmall, fill=WHITE
            )
            i += 1

    def render_display(self) -> None:
        """Render display."""
        data = self._host_data.get(self._current_screen)
        if data:
            with canvas(self._device) as draw:
                if self._output_groups and self._current_screen in self._output_groups:
                    self._draw_output(data, draw)
                elif self._current_screen == UPTIME:
                    self._draw_uptime(data, draw)
                else:
                    self._draw_standard(data, draw)
        if not self._sleep_handle and self._sleep_timeout.total_seconds > 0:
            self._sleep_handle = async_track_point_in_time(
                loop=self._loop,
                action=lambda x: self._sleeptime(),
                point_in_time=utcnow() + self._sleep_timeout.as_timedelta,
            )

    def handle_data_update(self, type: str):
        """Callback to handle new data present into screen."""
        if type == self._current_screen and not self._sleep:
            self.render_display()

    def _handle_press(self, pin: any) -> None:
        """Handle press of PIN for OLED display."""
        if self._sleep_handle:
            self._sleep_handle()
            self._sleep_handle = None
        if not self._sleep:
            self._current_screen = next(self._screen_order)
        else:
            self._sleep = False
        self.render_display()



"""Dallas temp sensor."""

import asyncio
import logging
from datetime import datetime
from ast import literal_eval

from adafruit_ds18x20 import DS18X20
from w1thermsensor import SensorNotReadyError

from boneio.const import SENSOR, STATE, TEMPERATURE
from boneio.helper import AsyncUpdater, BasicMqtt
from boneio.helper.exceptions import OneWireError
from boneio.helper.onewire import AsyncBoneIOW1ThermSensor, OneWireAddress, OneWireBus

from . import TempSensor

_LOGGER = logging.getLogger(__name__)


class DallasSensorDS2482(TempSensor, AsyncUpdater):
    DefaultName = TEMPERATURE
    SensorClass = DS18X20

    def __init__(
        self,
        bus: OneWireBus,
        address: OneWireAddress,
        id: str = DefaultName,
        **kwargs,
    ):
        """Initialize Temp class."""
        self._loop = asyncio.get_event_loop()
        BasicMqtt.__init__(self, id=id, topic_type=SENSOR, **kwargs)
        try:
            self._pct = DS18X20(bus=bus, address=address)
            self._state = None
        except ValueError as err:
            raise OneWireError(err)
        AsyncUpdater.__init__(self, **kwargs)


class DallasSensorW1(TempSensor, AsyncUpdater):
    DefaultName = TEMPERATURE
    SensorClass = AsyncBoneIOW1ThermSensor

    def __init__(
        self,
        address: OneWireAddress,
        id: str = DefaultName,
        filters: list = ["round(x, 2)"],
        **kwargs,
    ):
        """Initialize Temp class."""
        self._loop = asyncio.get_event_loop()
        BasicMqtt.__init__(self, id=id, topic_type=SENSOR, **kwargs)
        self._filters = filters
        try:
            self._pct = AsyncBoneIOW1ThermSensor(sensor_id=address)
        except ValueError as err:
            raise OneWireError(err)
        AsyncUpdater.__init__(self, **kwargs)

    async def async_update(self, time: datetime) -> None:
        try:
            _temp = await self._pct.get_temperature()
            _LOGGER.debug("Fetched temperature %s. Applying filters.", _temp)
            _temp = self._apply_filters(value=_temp)
            if not _temp:
                return
            self._state = _temp
            self._send_message(
                topic=self._send_topic,
                payload={STATE: self._state},
            )
        except SensorNotReadyError as err:
            _LOGGER.error("Sensor not ready, can't update %s", err)



"""GPIOInputButton to receive signals."""
import logging
from functools import partial

from boneio.const import BOTH, PRESSED, RELEASED
from boneio.helper import GpioBaseClass, edge_detect

_LOGGER = logging.getLogger(__name__)


class GpioInputSensor(GpioBaseClass):
    """Represent Gpio sensor on input boards."""

    def __init__(self, **kwargs) -> None:
        """Setup GPIO Input Button"""
        super().__init__(**kwargs)
        self._click_type = (
            (PRESSED, RELEASED)
            if kwargs.get("inverted", False)
            else (PRESSED, RELEASED)
        )
        edge_detect(
            self._pin,
            callback=self._handle_press,
            bounce=self._bounce_time.total_milliseconds,
            edge=BOTH,
        )
        _LOGGER.debug("Configured sensor pin %s", self._pin)

    def _handle_press(self, pin: str) -> None:
        """Handle the button press callback"""
        # Ignore if we are in a long press
        click_type = self._click_type[0] if self.is_pressed else self._click_type[1]
        _LOGGER.debug("%s event on pin %s", click_type, self._pin)
        self._loop.call_soon_threadsafe(
            partial(self._press_callback, click_type, self._pin)
        )



# from typing import Literal
try:
    from Adafruit_BBIO.GPIO import BOTH, FALLING, HIGH, LOW, RISING
except ModuleNotFoundError:
    HIGH = "high"
    LOW = "low"
    BOTH = "both"
    FALLING = "falling"
    RISING = "rising"
    pass
from typing_extensions import Literal

BONEIO = "boneIO"
NONE = "none"

# MISCELLANEOUS CONSTS
RELAY = "relay"
ON = "ON"
OFF = "OFF"
STATE = "state"
ENABLED = "enabled"
OUTPUT = "output"
PIN = "pin"
ID = "id"
KIND = "kind"
GPIO = "gpio"
GPIO_MODE = "gpio_mode"
ACTIONS = "actions"
ACTION = "action"
OUTPUT = "output"
SWITCH = "switch"
LIGHT = "light"
BUTTON = "button"
CONFIG_PIN = "/usr/bin/config-pin"
UPDATE_INTERVAL = "update_interval"
ADC = "adc"
IP = "ip"
MASK = "mask"
MAC = "mac"
NONE = "none"
MODBUS = "modbus"
UART = "uart"
RX = "rx"
TX = "tx"
RESTORE_STATE = "restore_state"
MODEL = "model"
UARTS = {
    "uart1": {ID: "/dev/ttyS1", TX: "P9.24", RX: "P9.26"},
    "uart2": {ID: "/dev/ttyS2", TX: "P9.21", RX: "P9.22"},
    "uart3": {ID: "/dev/ttyS3", TX: "P9.42", RX: None},
    "uart4": {ID: "/dev/ttyS4", TX: "P9.13", RX: "P9.11"},
    "uart5": {ID: "/dev/ttyS5", TX: "P8.37", RX: "P8.38"},
}

relay_actions = {ON: "turn_on", OFF: "turn_off", "TOGGLE": "toggle"}

# HA CONSTS
HOMEASSISTANT = "homeassistant"
HA_DISCOVERY = "ha_discovery"
OUTPUT_TYPE = "output_type"
SHOW_HA = "show_in_ha"

# OLED CONST
OLED = "oled"
FONTS = "fonts"
OLED_PIN = "P9_41"
GIGABYTE = 1073741824
MEGABYTE = 1048576
WIDTH = 128
UPTIME = "uptime"
NETWORK = "network"
CPU = "cpu"
DISK = "disk"
MEMORY = "memory"
SWAP = "swap"
WHITE = "white"

# INPUT CONST
INPUT = "input"
SINGLE = "single"
DOUBLE = "double"
LONG = "long"
PRESSED = "pressed"
RELEASED = "released"


# MQTT CONST
PAHO = "paho.mqtt.client"
PYMODBUS = "pymodbus"
MQTT = "mqtt"
HOST = "host"
USERNAME = "username"
PASSWORD = "password"
PORT = "port"
ONLINE = "online"
OFFLINE = "offline"
TOPIC = "topic"
TOPIC_PREFIX = "topic_prefix"

# I2C and MCP CONST
ADDRESS = "address"
MCP23017 = "mcp23017"
MCP = "mcp"
MCP_ID = "mcp_id"
INIT_SLEEP = "init_sleep"

# SENSOR CONST
TEMPERATURE = "temperature"
SENSOR = "sensor"
BINARY_SENSOR = "binary_sensor"
LM75 = "lm75"
MCP_TEMP_9808 = "mcp9808"
INPUT_SENSOR = "inputsensor"
DS2482 = "ds2482"
DALLAS = "dallas"
ONEWIRE = "onewire"

BASE = "base"
LENGTH = "length"
REGISTERS = "registers"

COVER = "cover"
IDLE = "idle"
OPENING = "opening"
CLOSING = "closing"
CLOSED = "closed"
OPEN = "open"
CLOSE = "close"
STOP = "stop"

# TYPING
ClickTypes = Literal[SINGLE, DOUBLE, LONG, PRESSED, RELEASED]
OledDataTypes = Literal[UPTIME, NETWORK, CPU, DISK, MEMORY, SWAP, OUTPUT]
Gpio_States = Literal[HIGH, LOW]
Gpio_Edges = Literal[BOTH, FALLING]
InputTypes = Literal[INPUT, INPUT_SENSOR]
DEVICE_CLASS = "device_class"
DallasBusTypes = Literal[DS2482, DALLAS]
FILTERS = "filters"



"""Sensor module."""
from boneio.sensor.adc import GpioADCSensor, initialize_adc
from boneio.sensor.gpio import GpioInputSensor
from boneio.sensor.temp.dallas import DallasSensorDS2482
from boneio.sensor.temp.lm75 import LM75Sensor
from boneio.sensor.temp.mcp9808 import MCP9808Sensor

__all__ = [
    "DallasSensorDS2482",
    "LM75Sensor",
    "MCP9808Sensor",
    "GpioInputSensor",
    "initialize_adc",
    "GpioADCSensor",
]



"""LM75 temp sensor."""
from adafruit_pct2075 import PCT2075

from boneio.const import LM75

from . import TempSensor


class LM75Sensor(TempSensor):
    """Represent LM75 sensor in BoneIO."""

    SensorClass = PCT2075
    DefaultName = LM75



import asyncio
import logging

try:
    from Adafruit_BBIO import GPIO
except ModuleNotFoundError:

    class GPIO:
        PUD_OFF = "poff"
        PUD_UP = "poff"
        PUD_DOWN = "poff"

        def __init__(self):
            pass

    pass

import subprocess
from typing import Callable

from boneio.const import CONFIG_PIN, FALLING
from boneio.const import GPIO as GPIO_STR
from boneio.const import GPIO_MODE, LOW, ClickTypes, Gpio_Edges, Gpio_States
from boneio.helper.exceptions import GPIOInputException
from boneio.helper.timeperiod import TimePeriod

_LOGGER = logging.getLogger(__name__)


def configure_pin(pin: str, mode: str = GPIO_STR) -> None:
    pin = f"{pin[0:3]}0{pin[3]}" if len(pin) == 4 else pin
    _LOGGER.debug(f"Configuring pin {pin} for mode {mode}.")
    subprocess.run(
        [CONFIG_PIN, pin, mode],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.STDOUT,
        timeout=1,
    )


def setup_output(pin: str) -> None:
    """Set up a GPIO as output."""

    GPIO.setup(pin, GPIO.OUT, pull_up_down=GPIO.PUD_DOWN)


gpio_modes = {
    "gpio": GPIO.PUD_OFF,
    "gpio_pu": GPIO.PUD_UP,
    "gpio_pd": GPIO.PUD_DOWN,
    "gpio_input": GPIO.PUD_OFF,
}


def setup_input(pin: str, pull_mode: str = "gpio") -> None:
    """Set up a GPIO as input."""
    gpio_mode = gpio_modes.get(pull_mode, GPIO.PUD_OFF)
    try:
        GPIO.setup(pin, GPIO.IN, gpio_mode)
    except (ValueError, SystemError) as err:
        raise GPIOInputException(err)


def write_output(pin: str, value: str) -> None:
    """Write a value to a GPIO."""

    GPIO.output(pin, value)


def read_input(pin: str, on_state: Gpio_States = LOW) -> None:
    """Read a value from a GPIO."""
    return GPIO.input(pin) is on_state


def edge_detect(
    pin: str, callback: Callable, bounce: int = 0, edge: Gpio_Edges = FALLING
) -> None:
    """Add detection for RISING and FALLING events."""
    try:
        GPIO.add_event_detect(gpio=pin, edge=edge, callback=callback, bouncetime=bounce)
    except RuntimeError as err:
        raise GPIOInputException(err)


class GpioBaseClass:
    """Base class for initialize GPIO"""

    def __init__(
        self, pin: str, press_callback: Callable[[ClickTypes, str], None], **kwargs
    ) -> None:
        """Setup GPIO Input Button"""
        self._pin = pin
        gpio_mode = kwargs.get(GPIO_MODE, GPIO_STR)
        self._bounce_time = kwargs.get("bounce_time", TimePeriod(milliseconds=10))
        self._loop = asyncio.get_running_loop()
        self._press_callback = press_callback
        setup_input(pin=self._pin, pull_mode=gpio_mode)

    @property
    def is_pressed(self):
        """Is button pressed."""
        return read_input(self._pin)



"""OneWire base class for DS2482.
Module by https://github.com/fgervais/ds2482
"""
import adafruit_onewire.bus
from onewireio import OneWire as OneWireIO
from boneio.helper.onewire import DS2482
from typing import List
from adafruit_onewire.bus import OneWireAddress as AdafruitOneWireAddress


def ds_address(rom: bytes) -> int:
    return int.from_bytes(rom, "little")


def reverse_dallas_id(a: str) -> str:
    """Reverse sensor address."""
    return "".join(reversed([a[i : i + 2] for i in range(0, len(a), 2)]))


class OneWireAddress(AdafruitOneWireAddress):
    @property
    def int_address(self) -> int:
        return ds_address(self.rom)

    @property
    def hex_id(self) -> str:
        return reverse_dallas_id(self.rom.hex())

    @property
    def hw_id(self) -> str:
        return self.hex_id[2:-2]


class OneWire(OneWireIO):
    def __init__(self, ds2482: DS2482):
        self.ds2482 = ds2482

    def deinit(self):
        """
        Deinitialize the OneWire bus and release any hardware resources for reuse.
        """
        self.ds2482.device_reset()

    def reset(self):
        """
        Reset the OneWire bus and read presence

        :return: False when at least one device is present
        """
        return self.ds2482.reset()

    def read_bit(self):
        """
        Read in a bit
        """
        return self.ds2482.single_bit()

    def write_bit(self, value):
        """
        Write out a bit based on value.
        """
        self.ds2482.single_bit(value)


class OneWireBus(adafruit_onewire.bus.OneWireBus):
    def __init__(self, ds2482: DS2482):
        self._ow = OneWire(ds2482)
        self._readbit = self._ow.read_bit
        self._writebit = self._ow.write_bit
        self._maximum_devices = adafruit_onewire.bus._MAX_DEV

    def scan(self) -> List[OneWireAddress]:
        """Scan for devices on the bus and return a list of addresses."""
        devices = []
        diff = 65
        rom = None
        count = 0
        for _ in range(0xFF):
            rom, diff = self._search_rom(rom, diff)
            if rom:
                count += 1
                if count > self.maximum_devices:
                    raise RuntimeError(
                        "Maximum device count of {} exceeded.".format(
                            self.maximum_devices
                        )
                    )
                devices.append(OneWireAddress(rom))
            if diff == 0:
                break
        return devices



from __future__ import annotations
import asyncio

# Typing imports that create a circular dependency
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..manager import Manager
from boneio.helper.timeperiod import TimePeriod


class AsyncUpdater:
    def __init__(self, manager: Manager, update_interval: TimePeriod, **kwargs):
        self.manager = manager
        self._update_interval = update_interval or TimePeriod(seconds=60)
        self.manager.append_task(coro=self._refresh, name=self.id)

    async def _refresh(self):
        while True:
            if hasattr(self, "async_update"):
                update_interval = (
                    await self.async_update(time=None)
                    or self._update_interval.total_in_seconds
                )
            else:
                update_interval = (
                    self.update(time=None) or self._update_interval.total_in_seconds
                )
            await asyncio.sleep(update_interval)



from __future__ import annotations
import asyncio
from datetime import datetime
import socket
import time
from math import floor
from typing import Callable, List
from functools import partial

import psutil

from boneio.const import (
    CPU,
    DISK,
    GIGABYTE,
    IP,
    MAC,
    MASK,
    MEGABYTE,
    MEMORY,
    NETWORK,
    NONE,
    SWAP,
    UPTIME,
    HOST,
)

# Typing imports that create a circular dependency
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from boneio.manager import Manager

from boneio.helper.async_updater import AsyncUpdater
from boneio.helper.timeperiod import TimePeriod
from boneio.sensor import LM75Sensor, MCP9808Sensor
from boneio.version import __version__

intervals = (("d", 86400), ("h", 3600), ("m", 60))


def display_time(seconds):
    """Strf time."""
    result = []

    for name, count in intervals:
        value = seconds // count
        if value:
            seconds -= value * count
            if value == 1:
                name = name.rstrip("s")
            result.append(f"{int(value)}{name}")
    return "".join(result)


def get_network_info():
    """Fetch network info."""

    def retrieve_from_psutil():
        addrs = psutil.net_if_addrs()["eth0"]
        out = {IP: NONE, MASK: NONE, MAC: NONE}
        for addr in addrs:
            if addr.family == socket.AF_INET:
                out["ip"] = addr.address
                out["mask"] = addr.netmask
            elif addr.family == psutil.AF_LINK:
                out["mac"] = addr.address
        return out

    return retrieve_from_psutil()


def get_cpu_info():
    """Fetch CPU info."""
    cpu = psutil.cpu_times_percent()
    return {
        "total": f"{int(100 - cpu.idle)}%",
        "user": f"{cpu.user}%",
        "system": f"{cpu.system}%",
    }


def get_disk_info():
    """Fetch disk info."""
    disk = psutil.disk_usage("/")
    return {
        "total": f"{floor(disk.total / GIGABYTE)}GB",
        "used": f"{floor(disk.used / GIGABYTE)}GB",
        "free": f"{floor(disk.free / GIGABYTE)}GB",
    }


def get_memory_info():
    """Fetch memory info."""
    vm = psutil.virtual_memory()
    return {
        "total": f"{floor(vm.total / MEGABYTE)}MB",
        "used": f"{floor(vm.used / MEGABYTE)}MB",
        "free": f"{floor(vm.available / MEGABYTE)}MB",
    }


def get_swap_info():
    """Fetch swap info."""
    swap = psutil.swap_memory()
    return {
        "total": f"{floor(swap.total / MEGABYTE)}MB",
        "used": f"{floor(swap.used / MEGABYTE)}MB",
        "free": f"{floor(swap.free / MEGABYTE)}MB",
    }


def get_uptime():
    """Fetch uptime info."""
    return display_time(time.clock_gettime(time.CLOCK_BOOTTIME))


class HostSensor(AsyncUpdater):
    """Host sensor."""

    def __init__(
        self,
        update_function: Callable,
        manager_callback: Callable,
        static_data: dict,
        id: str,
        type: str,
        **kwargs,
    ) -> None:
        self._update_function = update_function
        self._static_data = static_data
        self._state = None
        self._type = type
        self._manager_callback = manager_callback
        self._loop = asyncio.get_event_loop()
        self.id = id
        super().__init__(**kwargs)

    async def async_update(self, time: datetime) -> None:
        self._state = self._update_function()
        self._loop.call_soon_threadsafe(partial(self._manager_callback, self._type))

    @property
    def state(self) -> dict:
        if self._static_data:
            return {**self._static_data, self._type: self._state}
        return self._state


class HostData:
    """Helper class to store host data."""

    data = {UPTIME: {}, NETWORK: {}, CPU: {}, DISK: {}, MEMORY: {}, SWAP: {}}

    def __init__(
        self,
        output: dict,
        callback: Callable,
        temp_sensor: Callable[[LM75Sensor, MCP9808Sensor], None],
        manager: Manager,
        enabled_screens: List[str],
    ) -> None:
        """Initialize HostData."""
        self._hostname = socket.gethostname()
        host_stats = {
            NETWORK: {"f": get_network_info, "update_interval": TimePeriod(seconds=60)},
            CPU: {"f": get_cpu_info, "update_interval": TimePeriod(seconds=5)},
            DISK: {"f": get_disk_info, "update_interval": TimePeriod(seconds=60)},
            MEMORY: {"f": get_memory_info, "update_interval": TimePeriod(seconds=10)},
            SWAP: {"f": get_swap_info, "update_interval": TimePeriod(seconds=60)},
            UPTIME: {
                "f": get_uptime,
                "static": {HOST: self._hostname, "version": __version__},
                "update_interval": TimePeriod(seconds=30),
            },
        }
        for k, _v in host_stats.items():
            if k not in enabled_screens:
                continue
            self.data[k] = HostSensor(
                update_function=_v["f"],
                static_data=_v.get("static"),
                manager=manager,
                manager_callback=callback,
                id=f"{k}_hoststats",
                type=k,
                update_interval=_v["update_interval"],
            )
        self._temp_sensor = temp_sensor
        self._output = output
        self._callback = callback
        self._loop = asyncio.get_running_loop()

    def get(self, type: str) -> dict:
        """Get saved stats."""
        if type in self._output:
            return self._get_output(type)
        return self.data[type].state

    def _get_output(self, type: str) -> dict:
        """Get stats for output."""
        out = {}
        for output in self._output[type].values():
            out[output.id] = output.state
        return out



import logging

from boneio.const import PAHO, PYMODBUS
from boneio.version import __version__

_LOGGER = logging.getLogger(__name__)
_nameToLevel = {
    "CRITICAL": logging.CRITICAL,
    "FATAL": logging.FATAL,
    "ERROR": logging.ERROR,
    "WARN": logging.WARNING,
    "WARNING": logging.WARNING,
    "INFO": logging.INFO,
    "DEBUG": logging.DEBUG,
    "NOTSET": logging.NOTSET,
}


def configure_logger(log_config: dict, debug: int) -> None:
    """Configure logger based on config yaml."""

    def debug_logger():
        if debug == 0:
            logging.getLogger().setLevel(logging.INFO)
        if debug > 0:
            logging.getLogger().setLevel(logging.DEBUG)
            logging.getLogger(PAHO).setLevel(logging.WARN)
            logging.getLogger(PYMODBUS).setLevel(logging.WARN)
            logging.getLogger("pymodbus.client").setLevel(logging.WARN)
            _LOGGER.info("Debug mode active")
            _LOGGER.debug(f"Lib version is {__version__}")
        if debug > 1:
            logging.getLogger(PAHO).setLevel(logging.DEBUG)
            logging.getLogger(PYMODBUS).setLevel(logging.DEBUG)
            logging.getLogger("pymodbus.client").setLevel(logging.DEBUG)

    if not log_config:
        debug_logger()
        return
    default = log_config.get("default", "").upper()
    if default in _nameToLevel:
        _LOGGER.info("Setting default log level to %s", default)
        logging.getLogger().setLevel(_nameToLevel[default])
        if debug == 0:
            debug = -1
    for k, v in log_config.get("logs", {}).items():
        logger = logging.getLogger(k)
        val = v.upper()
        if val in _nameToLevel and logger:
            _LOGGER.info("Setting %s log level to %s", k, val)
            logger.setLevel(_nameToLevel[val])
    debug_logger()



"""Helper dir for BoneIO."""

from boneio.helper.exceptions import GPIOInputException, I2CError
from boneio.helper.gpio import (
    GpioBaseClass,
    configure_pin,
    edge_detect,
    read_input,
    setup_input,
    setup_output,
    write_output,
)
from boneio.helper.ha_discovery import (
    ha_adc_sensor_availabilty_message,
    ha_binary_sensor_availabilty_message,
    ha_button_availabilty_message,
    ha_input_availabilty_message,
    ha_light_availabilty_message,
    ha_sensor_availabilty_message,
    ha_sensor_temp_availabilty_message,
    ha_switch_availabilty_message,
)
from boneio.helper.mqtt import BasicMqtt
from boneio.helper.async_updater import AsyncUpdater
from boneio.helper.oled import make_font
from boneio.helper.queue import UniqueQueue
from boneio.helper.state_manager import StateManager
from boneio.helper.stats import HostData
from boneio.helper.timeperiod import TimePeriod
from boneio.helper.yaml_util import (
    CustomValidator,
    load_config_from_file,
    load_config_from_string,
    load_yaml_file,
    schema_file,
)

__all__ = [
    "CustomValidator",
    "load_yaml_file",
    "HostData",
    "setup_input",
    "setup_output",
    "edge_detect",
    "read_input",
    "write_output",
    "make_font",
    "ha_light_availabilty_message",
    "ha_switch_availabilty_message",
    "ha_sensor_availabilty_message",
    "ha_adc_sensor_availabilty_message",
    "ha_sensor_temp_availabilty_message",
    "ha_binary_sensor_availabilty_message",
    "ha_button_availabilty_message",
    "ha_input_availabilty_message",
    "GPIOInputException",
    "I2CError",
    "GpioBaseClass",
    "StateManager",
    "configure_pin",
    "BasicMqtt",
    "AsyncUpdater",
    "UniqueQueue",
    "schema_file",
    "load_config_from_string",
    "load_config_from_file",
    "TimePeriod",
    "callback",
    "is_callback",
]



"""State files manager."""
import asyncio
import json
from typing import Any


class StateManager:
    """StateManager to load and save states to file."""

    def __init__(self, state_file: str) -> None:
        """Initialize disk StateManager."""
        self._loop = asyncio.get_event_loop()
        self._lock = asyncio.Lock()
        self._file = state_file
        self._state = self.load_states()
        self._file_uptodate = False

    def load_states(self) -> dict:
        """Load state file."""
        try:
            with open(self._file, "r") as state_file:
                datastore = json.load(state_file)
                return datastore
        except FileNotFoundError:
            pass
        return {}

    def del_attribute(self, attr_type: str, attribute: str) -> None:
        """Delete attribute"""
        if attr_type in self._state and attribute in self._state[attr_type]:
            del self._state[attr_type][attribute]

    def save_attribute(self, attr_type: str, attribute: str, value: str) -> None:
        """Save single attribute to file."""
        if attr_type not in self._state:
            self._state[attr_type] = {}
        self._state[attr_type][attribute] = value
        asyncio.run_coroutine_threadsafe(self.save_state(), self._loop)

    def get(self, attr_type: str, attr: str, default_value: Any = None) -> Any:
        """Retrieve attribute from json."""
        attrs = self._state.get(attr_type)
        if attrs:
            return attrs.get(attr, default_value)
        return default_value

    @property
    def state(self) -> dict:
        """Retrieve all states."""
        return self._state

    async def save_state(self) -> None:
        """Async save state."""
        if self._lock.locked():
            # Let's not save state if something happens same time.
            return
        async with self._lock:
            with open(self._file, "w+", encoding="utf-8") as f:
                json.dump(self._state, f, indent=2)



import asyncio
import logging
from collections import deque
from typing import Callable, Coroutine, List, Optional, Set, Union
from board import SCL, SDA
from busio import I2C

from boneio.const import (
    ACTION,
    ADDRESS,
    BUTTON,
    CLOSE,
    COVER,
    DALLAS,
    ID,
    INPUT,
    LM75,
    MCP_TEMP_9808,
    MODBUS,
    MQTT,
    NONE,
    ONEWIRE,
    ONLINE,
    OPEN,
    OUTPUT,
    PIN,
    RELAY,
    STATE,
    STOP,
    TOPIC,
    UART,
    UARTS,
    ClickTypes,
    InputTypes,
    relay_actions,
    DS2482,
)
from boneio.helper import (
    GPIOInputException,
    HostData,
    I2CError,
    StateManager,
    ha_button_availabilty_message,
    ha_light_availabilty_message,
    ha_switch_availabilty_message,
)
from boneio.helper.config import ConfigHelper
from boneio.helper.events import EventBus
from boneio.helper.exceptions import ModbusUartException
from boneio.helper.loader import (
    configure_cover,
    configure_input,
    configure_relay,
    create_dallas_sensor,
    create_mcp23017,
    create_temp_sensor,
)
from boneio.helper.logger import configure_logger
from boneio.helper.yaml_util import load_config_from_file
from boneio.modbus import Modbus

from w1thermsensor.errors import KernelModuleLoadError

_LOGGER = logging.getLogger(__name__)


class Manager:
    """Manager to communicate MQTT with GPIO inputs and outputs."""

    def __init__(
        self,
        send_message: Callable[[str, Union[str, dict], bool], None],
        state_manager: StateManager,
        config_helper: ConfigHelper,
        config_file_path: str,
        relay_pins: List = [],
        input_pins: List = [],
        sensors: dict = {},
        modbus: dict = None,
        mcp23017: Optional[List] = None,
        ds2482: Optional[List] = [],
        dallas: Optional[dict] = None,
        oled: dict = {},
        adc_list: Optional[List] = None,
        covers: Optional[List] = [],
    ) -> None:
        """Initialize the manager."""
        _LOGGER.info("Initializing manager module.")

        self._loop = asyncio.get_event_loop()
        self._config_helper = config_helper
        self._host_data = None
        self._config_file_path = config_file_path
        self._state_manager = state_manager
        self._event_bus = EventBus(self._loop)
        self._autodiscovery_messages = []

        self.send_message = send_message
        self._input_pins = input_pins
        self._i2cbusio = I2C(SCL, SDA)
        self._mcp = {}
        self._output = {}
        self._oled = None
        self._tasks: List[asyncio.Task] = []
        self._covers = {}
        self._temp_sensors = []
        self._modbus = None

        self._configure_modbus(modbus=modbus)

        self._configure_temp_sensors(sensors=sensors)

        self._configure_modbus_sensors(sensors=sensors)
        self._configure_sensors(
            dallas=dallas, ds2482=ds2482, sensors=sensors.get(ONEWIRE)
        )

        self.grouped_outputs = create_mcp23017(
            manager=self, mcp23017=mcp23017, i2cbusio=self._i2cbusio
        )

        self._configure_adc(adc_list=adc_list)

        for _config in relay_pins:
            _id = _config[ID].replace(" ", "")
            out = configure_relay(
                manager=self,
                state_manager=self._state_manager,
                topic_prefix=self._config_helper.topic_prefix,
                relay_id=_id,
                relay_callback=self._relay_callback,
                config=_config,
            )
            if not out:
                continue
            self._output[_id] = out
            if out.output_type != NONE:
                self.send_ha_autodiscovery(
                    id=out.id,
                    name=out.name,
                    ha_type=out.output_type,
                    availability_msg_func=ha_light_availabilty_message
                    if out.is_light
                    else ha_switch_availabilty_message,
                )
            self._loop.call_soon_threadsafe(
                self._loop.call_later,
                0.5,
                out.send_state,
            )

        for _config in covers:
            _id = _config[ID].replace(" ", "")
            open_relay = self._output.get(_config.get("open_relay"))
            close_relay = self._output.get(_config.get("close_relay"))
            if not open_relay:
                _LOGGER.error(
                    "Can't configure cover %s. This relay doesn't exist.",
                    _config.get("open_relay"),
                )
                continue
            if not close_relay:
                _LOGGER.error(
                    "Can't configure cover %s. This relay doesn't exist.",
                    _config.get("close_relay"),
                )
                continue
            if open_relay.output_type != NONE or close_relay.output_type != NONE:
                _LOGGER.error(
                    "Can't configure cover %s. %s",
                    _id,
                    "You have to explicity set types of relays to None so you can't turn it on directly.",
                )
                continue
            self._covers[_id] = configure_cover(
                manager=self,
                cover_id=_id,
                state_manager=self._state_manager,
                config=_config,
                open_relay=open_relay,
                close_relay=close_relay,
                open_time=_config.get("open_time"),
                close_time=_config.get("close_time"),
                event_bus=self._event_bus,
                send_ha_autodiscovery=self.send_ha_autodiscovery,
                topic_prefix=self._config_helper.topic_prefix,
            )

        _LOGGER.info("Initializing inputs. This will take a while.")
        input_pins = set()
        for gpio in self._input_pins:
            pin = gpio[PIN]
            if pin in input_pins:
                _LOGGER.warn("This PIN %s is already configured. Omitting it.", pin)
                continue
            input_pins.add(
                configure_input(
                    gpio=gpio,
                    pin=pin,
                    press_callback=self.press_callback,
                    send_ha_autodiscovery=self.send_ha_autodiscovery,
                )
            )

        if oled.get("enabled", False):
            from boneio.oled import Oled

            screens = oled.get("screens")

            self._host_data = HostData(
                manager=self,
                enabled_screens=screens,
                output=self.grouped_outputs,
                temp_sensor=self._temp_sensors[0] if self._temp_sensors else None,
                callback=self._host_data_callback,
            )
            try:
                self._oled = Oled(
                    host_data=self._host_data,
                    screen_order=screens,
                    output_groups=list(self.grouped_outputs),
                    sleep_timeout=oled.get("screensaver_timeout", 60),
                )
            except (GPIOInputException, I2CError) as err:
                _LOGGER.error("Can't configure OLED display. %s", err)
        self.prepare_ha_buttons()

        _LOGGER.info("BoneIO manager is ready.")

    def append_task(self, coro: Coroutine, name: str = "Unknown") -> asyncio.Future:
        """Add task to run with asyncio loop."""
        _LOGGER.debug("Appending update task for %s", name)
        task: asyncio.Future = asyncio.create_task(coro())
        self._tasks.append(task)
        return task

    def _configure_sensors(
        self, dallas: Optional[dict], ds2482: Optional[List], sensors: Optional[List]
    ):
        """
        Configure Dallas sensors via GPIO PIN bus or DS2482 bus.
        """
        if not ds2482 and not dallas:
            return
        from boneio.helper.loader import (
            find_onewire_devices,
        )

        _one_wire_devices = {}
        _ds_onewire_bus = {}

        for _single_ds in ds2482:
            _LOGGER.debug("Preparing DS2482 bus at address %s.", _single_ds[ADDRESS])
            from boneio.helper.loader import (
                configure_ds2482,
            )
            from boneio.sensor import DallasSensorDS2482

            _ds_onewire_bus[_single_ds[ID]] = configure_ds2482(
                i2cbusio=self._i2cbusio, address=_single_ds[ADDRESS]
            )
            _one_wire_devices.update(
                find_onewire_devices(
                    ow_bus=_ds_onewire_bus[_single_ds[ID]],
                    bus_id=_single_ds[ID],
                    bus_type=DS2482,
                )
            )
        if dallas:
            _LOGGER.debug("Preparing Dallas bus.")
            from boneio.helper.loader import configure_dallas

            try:
                from w1thermsensor.kernel import load_kernel_modules

                load_kernel_modules()
                from boneio.sensor.temp.dallas_w1 import DallasSensorW1

                _one_wire_devices.update(
                    find_onewire_devices(
                        ow_bus=configure_dallas(),
                        bus_id=dallas[ID],
                        bus_type=DALLAS,
                    )
                )
            except KernelModuleLoadError as err:
                _LOGGER.error("Can't configure Dallas W1 device %s", err)
                pass

        for sensor in sensors:
            address = _one_wire_devices.get(sensor[ADDRESS])
            if not address:
                continue
            ds2482_bus_id = sensor.get("bus_id")
            if ds2482_bus_id and ds2482_bus_id in _ds_onewire_bus:
                kwargs = {
                    "bus": _ds_onewire_bus[ds2482_bus_id],
                    "cls": DallasSensorDS2482,
                }
            else:
                kwargs = {"cls": DallasSensorW1}
            _LOGGER.debug("Configuring sensor %s for boneIO", address)
            self._temp_sensors.append(
                create_dallas_sensor(
                    manager=self,
                    address=address,
                    topic_prefix=self._config_helper.topic_prefix,
                    config=sensor,
                    **kwargs,
                )
            )

    def _configure_adc(self, adc_list: Optional[List]) -> None:
        if adc_list:
            from boneio.helper.loader import create_adc

            create_adc(
                manager=self,
                topic_prefix=self._config_helper.topic_prefix,
                adc_list=adc_list,
            )

    def _configure_modbus(self, modbus: dict) -> None:
        if modbus and modbus.get(UART) in UARTS:
            try:
                self._modbus = Modbus(UARTS[modbus.get(UART)])
            except ModbusUartException:
                _LOGGER.error(
                    "This UART %s can't be used for modbus communication.",
                    modbus.get(UART),
                )
                self._modbus = None

    def _configure_temp_sensors(self, sensors: dict) -> None:
        for sensor_type in (LM75, MCP_TEMP_9808):
            if sensors.get(sensor_type):
                for temp_def in sensors.get(sensor_type):
                    temp_sensor = create_temp_sensor(
                        manager=self,
                        topic_prefix=self._config_helper.topic_prefix,
                        sensor_type=sensor_type,
                        config=temp_def,
                        i2cbusio=self._i2cbusio,
                    )
                    if temp_sensor:
                        self._temp_sensors.append(temp_sensor)

    def _configure_modbus_sensors(self, sensors: dict) -> None:
        if sensors.get(MODBUS) and self._modbus:
            from boneio.helper.loader import create_modbus_sensors

            create_modbus_sensors(
                manager=self,
                event_bus=self._event_bus,
                sensors=sensors.get(MODBUS),
                modbus=self._modbus,
                config_helper=self._config_helper,
            )

    async def reconnect_callback(self) -> None:
        """Function to invoke when connection to MQTT is (re-)established."""
        _LOGGER.info("Sending online state.")
        topic = f"{self._config_helper.topic_prefix}/{STATE}"
        self.send_message(topic=topic, payload=ONLINE, retain=True)

    def _relay_callback(
        self, relay_type: str, relay_id: str, restore_state: bool
    ) -> None:
        """Relay callback function."""
        if restore_state:
            self._state_manager.save_attribute(
                attr_type=RELAY,
                attribute=relay_id,
                value=self._output[relay_id].is_active,
            )
        self._host_data_callback(type=relay_type)

    def _logger_reload(self) -> None:
        """_Logger reload function."""
        _config = load_config_from_file(config_file=self._config_file_path)
        if not _config:
            return
        configure_logger(log_config=_config.get("logger"), debug=-1)

    def _host_data_callback(self, type: str) -> None:
        if self._oled:
            self._oled.handle_data_update(type)

    def get_tasks(self) -> Set[asyncio.Task]:
        """Retrieve asyncio tasks to run."""
        return self._tasks

    def prepare_ha_buttons(self) -> None:
        """Prepare HA buttons for reload."""
        self.send_ha_autodiscovery(
            id="Logger",
            name="Logger reload",
            ha_type=BUTTON,
            availability_msg_func=ha_button_availabilty_message,
        )

    @property
    def mcp(self):
        """Get MCP by it's id."""
        return self._mcp

    def press_callback(
        self, x: ClickTypes, inpin: str, actions: List, input_type: InputTypes = INPUT
    ) -> None:
        """Press callback to use in input gpio.
        If relay input map is provided also toggle action on relay or cover or mqtt."""
        topic = f"{self._config_helper.topic_prefix}/{input_type}/{inpin}"
        self.send_message(topic=topic, payload=x)
        for action_definition in actions:
            _LOGGER.debug("Executing action %s", action_definition)
            if action_definition[ACTION] == OUTPUT:
                device = action_definition.get(PIN)
                if not device:
                    continue
                relay = self._output.get(device.replace(" ", ""))
                if relay:
                    getattr(relay, action_definition["action_output"])()
            elif action_definition[ACTION] == MQTT:
                action_topic = action_definition.get(TOPIC)
                action_payload = action_definition.get("action_mqtt_msg")
                if action_topic and action_payload:
                    self.send_message(topic=action_topic, payload=action_payload)
            elif action_definition[ACTION] == COVER:
                device = action_definition.get(PIN)
                if not device:
                    continue
                cover = self._covers.get(device.replace(" ", ""))
                if cover:
                    getattr(cover, action_definition["action_cover"])()

        # This is similar how Z2M is clearing click sensor.
        self._loop.call_soon_threadsafe(
            self._loop.call_later, 0.2, self.send_message, topic, ""
        )

    def send_ha_autodiscovery(
        self,
        id: str,
        name: str,
        ha_type: str,
        availability_msg_func: Callable,
        topic_prefix: str = None,
        **kwargs,
    ) -> None:
        """Send HA autodiscovery information for each relay."""
        if not self._config_helper.ha_discovery:
            return
        topic_prefix = topic_prefix or self._config_helper.topic_prefix
        payload = availability_msg_func(topic=topic_prefix, id=id, name=name, **kwargs)
        topic = f"{self._config_helper.ha_discovery_prefix}/{ha_type}/{topic_prefix}/{id}/config"
        _LOGGER.debug("Sending HA discovery for %s, %s.", ha_type, name)
        self._config_helper.add_autodiscovery_msg(topic=topic, payload=payload)
        self.send_message(topic=topic, payload=payload, retain=True)

    def resend_autodiscovery(self) -> None:
        for msg in self._config_helper.autodiscovery_msgs:
            self.send_message(**msg, retain=True)

    async def receive_message(self, topic: str, message: str) -> None:
        """Callback for receiving action from Mqtt."""
        _LOGGER.debug("Processing topic %s with message %s.", topic, message)
        if topic.startswith(f"{self._config_helper.ha_discovery_prefix}/status"):
            if message == ONLINE:
                self.resend_autodiscovery()
                self._event_bus.signal_ha_online()
            return
        assert topic.startswith(self._config_helper.cmd_topic_prefix)
        topic_parts_raw = topic[len(self._config_helper.cmd_topic_prefix) :].split("/")
        topic_parts = deque(topic_parts_raw)
        try:
            msg_type = topic_parts.popleft()
            device_id = topic_parts.popleft()
            command = topic_parts.pop()
            _LOGGER.debug(
                "Divide topic to: msg_type: %s, device_id: %s, command: %s",
                msg_type,
                device_id,
                command,
            )
        except IndexError:
            _LOGGER.error("Part of topic is missing. Not invoking command.")
            return

        if msg_type == RELAY and command == "set":
            target_device = self._output.get(device_id)

            if target_device and target_device.output_type != NONE:
                action_from_msg = relay_actions.get(message.upper())
                if action_from_msg:
                    getattr(target_device, action_from_msg)()
                else:
                    _LOGGER.debug("Action not exist %s.", message.upper())
            else:
                _LOGGER.debug("Target device not found %s.", device_id)
        elif msg_type == COVER:
            cover = self._covers.get(device_id)
            if not cover:
                return
            if command == "set":
                if message in (
                    OPEN,
                    CLOSE,
                    STOP,
                    "toggle",
                    "toggle_open",
                    "toggle_close",
                ):
                    getattr(cover, message.lower())()
            elif command == "pos":
                position = int(message)
                if 0 <= position <= 100:
                    await cover.set_cover_position(position=position)
                else:
                    _LOGGER.warn(
                        "Positon cannot be set. Not number between 0-100. %s", message
                    )
        elif msg_type == BUTTON and command == "set":
            if device_id == "logger" and message == "reload":
                _LOGGER.info("Reloading logger configuration.")
                self._logger_reload()

    @property
    def output(self) -> dict:
        """Get list of output."""
        return self._output



"""Basic Relay module."""

import asyncio
import logging
from typing import Callable

from boneio.const import LIGHT, NONE, OFF, ON, RELAY, STATE, SWITCH
from boneio.helper import BasicMqtt

_LOGGER = logging.getLogger(__name__)


class BasicRelay(BasicMqtt):
    """Basic relay class."""

    def __init__(
        self,
        callback: Callable,
        id: str = None,
        output_type=SWITCH,
        restored_state: bool = False,
        **kwargs,
    ) -> None:
        """Initialize Basic relay."""
        self._momentary_turn_on = kwargs.pop("momentary_turn_on", None)
        self._momentary_turn_off = kwargs.pop("momentary_turn_off", None)
        super().__init__(id=id, name=id, topic_type=RELAY, **kwargs)
        self._output_type = output_type
        if output_type == NONE:
            self._momentary_turn_on = None
            self._momentary_turn_off = None
        self._state = restored_state
        self._callback = callback
        self._loop = asyncio.get_running_loop()

    @property
    def is_mcp_type(self) -> bool:
        """Check if relay is mcp type."""
        return False

    @property
    def output_type(self) -> str:
        """HA type."""
        return self._output_type

    @property
    def is_light(self) -> bool:
        """Check if HA type is light"""
        return self._output_type == LIGHT

    @property
    def id(self) -> bool:
        """Id of the relay.
        Has to be trimmed out of spaces because of MQTT handling in HA."""
        return self._id or self._pin

    @property
    def name(self) -> bool:
        """Not trimmed id."""
        return self._name or self._pin

    @property
    def state(self) -> bool:
        """Is relay active."""
        return self._state

    def send_state(self) -> None:
        """Send state to Mqtt on action."""
        state = ON if self.is_active else OFF
        self._state = state
        if self.output_type != NONE:
            self._send_message(
                topic=self._send_topic,
                payload={STATE: state},
            )
        self._loop.call_soon_threadsafe(self._callback)

    def toggle(self) -> None:
        """Toggle relay."""
        _LOGGER.debug("Toggle relay.")
        if self.is_active:
            self.turn_off()
        else:
            self.turn_on()

    @property
    def is_active(self) -> bool:
        """Is active check."""
        raise NotImplementedError

    def turn_on(self) -> None:
        """Call turn on action."""
        raise NotImplementedError

    def turn_off(self) -> None:
        """Call turn off action."""
        raise NotImplementedError



"""BoneIO Errors"""


class BoneIOException(Exception):
    """BoneIO standard exception."""


class GPIOInputException(BoneIOException):
    """GPIOInput Exception."""


class I2CError(BoneIOException):
    """I2C Exception."""


class OneWireError(BoneIOException):
    """One Wire Exception."""


class ConfigurationException(BoneIOException):
    """Configuration yaml exception."""


class CoverRelayException(BoneIOException):
    """Cover configuration exception."""


class ModbusUartException(BoneIOException):
    """Cover configuration exception."""



"""Cover module."""
from __future__ import annotations
import asyncio
import logging
from typing import Callable

from boneio.const import CLOSE, CLOSED, CLOSING, COVER, IDLE, OPEN, OPENING, STOP
from boneio.helper.events import EventBus
from boneio.helper.mqtt import BasicMqtt
from boneio.helper.timeperiod import TimePeriod
from boneio.relay import MCPRelay

_LOGGER = logging.getLogger(__name__)

COVER_COMMANDS = {
    OPEN: "open_cover",
    CLOSE: "close_cover",
    STOP: "stop",
    "toggle": "toggle",
    "toggle_open": "toggle_open",
}


class RelayHelper:
    """Relay helper for cover either open/close."""

    def __init__(self, relay: MCPRelay, time: TimePeriod) -> None:
        """Initialize helper."""
        self._relay = relay
        self._steps = 100 / time.total_seconds

    @property
    def relay(self) -> MCPRelay:
        """Get relay."""
        return self._relay

    @property
    def steps(self) -> int:
        """Get steps for each time."""
        return self._steps


class Cover(BasicMqtt):
    """Cover class of boneIO"""

    def __init__(
        self,
        id: str,
        open_relay: MCPRelay,
        close_relay: MCPRelay,
        state_save: Callable,
        open_time: TimePeriod,
        close_time: TimePeriod,
        event_bus: EventBus,
        restored_state: int = 100,
        **kwargs,
    ) -> None:
        """Initialize cover class."""
        self._loop = asyncio.get_event_loop()
        self._id = id
        super().__init__(id=id, name=id, topic_type=COVER, **kwargs)
        self._lock = asyncio.Lock()
        self._state_save = state_save
        self._open = RelayHelper(relay=open_relay, time=open_time)
        self._close = RelayHelper(relay=close_relay, time=close_time)
        self._set_position = None
        self._current_operation = IDLE
        self._position = restored_state
        self._requested_closing = True
        self._event_bus = event_bus
        self._timer_handle = None
        if self._position is None:
            self._closed = True
        else:
            self._closed = self._position <= 0
        self._event_bus.add_sigterm_listener(self.on_exit)
        self._loop.call_soon_threadsafe(
            self._loop.call_later,
            0.5,
            self.send_state,
        )

    async def run_cover(
        self,
        current_operation: str,
    ) -> None:
        """Run cover engine."""
        if self._current_operation != IDLE:
            self._stop_cover()
        self._current_operation = current_operation

        def get_relays():
            if current_operation == OPENING:
                return (self._open.relay, self._close.relay)
            else:
                return (self._close.relay, self._open.relay)

        (relay, inverted_relay) = get_relays()
        async with self._lock:
            if inverted_relay.is_active:
                inverted_relay.turn_off()
            self._timer_handle = self._event_bus.add_listener(
                f"{COVER}{self.id}", self.listen_cover
            )
            relay.turn_on()

    def on_exit(self) -> None:
        """Stop on exit."""
        self._stop_cover(on_exit=True)

    @property
    def cover_state(self) -> CLOSED | OPEN:
        """Current state of cover."""
        return CLOSED if self._closed else OPEN

    def stop(self) -> None:
        """Public Stop cover graceful."""
        _LOGGER.info("Stopping cover %s.", self._id)
        if self._current_operation != IDLE:
            self._stop_cover(on_exit=False)

    def send_state(self) -> None:
        """Send state of cover to mqtt."""
        self._send_message(topic=f"{self._send_topic}/state", payload=self.cover_state)
        pos = round(self._position, 0)
        self._send_message(topic=f"{self._send_topic}/pos", payload=str(pos))
        self._state_save(position=pos)

    def _stop_cover(self, on_exit=False) -> None:
        """Stop cover."""
        self._open.relay.turn_off()
        self._close.relay.turn_off()
        if self._timer_handle is not None:
            self._event_bus.remove_listener(f"{COVER}{self.id}")
            self._timer_handle = None
            self._set_position = None
            if not on_exit:
                self.send_state()
        self._current_operation = IDLE

    @property
    def current_cover_position(self) -> int:
        """Return the current position of the cover."""
        return round(self._position, 0)

    def listen_cover(self, *args) -> None:
        """Listen for change in cover."""
        if self._current_operation == IDLE:
            return

        def get_step():
            """Get step for current operation."""
            if self._requested_closing:
                return -self._close.steps
            else:
                return self._open.steps

        step = get_step()
        self._position += step
        rounded_pos = round(self._position, 0)
        if self._set_position:
            # Set position is only working for every 10%, so round to nearest 10.
            # Except for start moving time
            if (self._requested_closing and rounded_pos < 95) or rounded_pos > 5:
                rounded_pos = round(self._position, -1)
        else:
            if rounded_pos > 100:
                rounded_pos = 100
            elif rounded_pos < 0:
                rounded_pos = 0
        self._send_message(topic=f"{self._send_topic}/pos", payload=rounded_pos)
        if rounded_pos == self._set_position or (
            self._set_position is None and (rounded_pos >= 100 or rounded_pos <= 0)
        ):
            self._position = rounded_pos
            self._closed = self.current_cover_position <= 0
            self._stop_cover()
            return

        self._closed = self.current_cover_position <= 0

    async def close_cover(self) -> None:
        """Close cover."""
        if self._position == 0:
            return
        if self._position is None:
            self._closed = True
            return
        _LOGGER.info("Closing cover %s.", self._id)

        self._requested_closing = True
        self._send_message(topic=f"{self._send_topic}/state", payload=CLOSING)
        await self.run_cover(
            current_operation=CLOSING,
        )

    async def open_cover(self) -> None:
        """Open cover."""
        if self._position == 100:
            return
        if self._position is None:
            self._closed = False
            return
        _LOGGER.info("Opening cover %s.", self._id)

        self._requested_closing = False
        self._send_message(topic=f"{self._send_topic}/state", payload=OPENING)
        await self.run_cover(
            current_operation=OPENING,
        )

    async def set_cover_position(self, position: int) -> None:
        """Move cover to a specific position."""
        set_position = round(position, -1)
        if self._position == position or set_position == self._set_position:
            return
        if self._set_position:
            self._stop_cover(on_exit=True)
        _LOGGER.info("Setting cover at position %s.", set_position)
        self._set_position = set_position

        self._requested_closing = set_position < self._position
        current_operation = CLOSING if self._requested_closing else OPENING
        _LOGGER.debug(
            "Requested set position %s. Operation %s", set_position, current_operation
        )
        self._send_message(topic=f"{self._send_topic}/state", payload=current_operation)
        await self.run_cover(
            current_operation=current_operation,
        )

    def open(self) -> None:
        _LOGGER.debug("Opening cover %s.", self._id)
        asyncio.create_task(self.open_cover())

    def close(self) -> None:
        _LOGGER.debug("Closing cover %s.", self._id)
        asyncio.create_task(self.close_cover())

    def toggle(self) -> None:
        _LOGGER.debug("Toggle cover %s from input.", self._id)
        if self.cover_state == CLOSED:
            self.close()
        else:
            self.open()

    def toggle_open(self) -> None:
        _LOGGER.debug("Toggle open cover %s from input.", self._id)
        if self._current_operation != IDLE:
            self.stop()
        else:
            self.open()

    def toggle_close(self) -> None:
        _LOGGER.debug("Toggle close cover %s from input.", self._id)
        if self._current_operation != IDLE:
            self.stop()
        else:
            self.close()



"""Bonecli script."""
import argparse
import asyncio
import logging
import sys
import os

os.environ["W1THERMSENSOR_NO_KERNEL_MODULE"] = "1"

from colorlog import ColoredFormatter
from yaml import MarkedYAMLError

from boneio.const import ACTION
from boneio.helper import load_config_from_file
from boneio.helper.exceptions import ConfigurationException
from boneio.helper.logger import configure_logger
from boneio.runner import async_run
from boneio.version import __version__

TASK_CANCELATION_TIMEOUT = 1

_LOGGER = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)
fmt = "%(asctime)s %(levelname)s (%(threadName)s) [%(name)s] %(message)s"
datefmt = "%Y-%m-%d %H:%M:%S"
colorfmt = f"%(log_color)s{fmt}%(reset)s"
logging.getLogger().handlers[0].setFormatter(
    ColoredFormatter(
        colorfmt,
        datefmt=datefmt,
        reset=True,
        log_colors={
            "DEBUG": "cyan",
            "INFO": "green",
            "WARNING": "yellow",
            "ERROR": "red",
            "CRITICAL": "red",
        },
    )
)


def get_arguments() -> argparse.Namespace:
    """Get parsed passed in arguments."""

    parser = argparse.ArgumentParser(
        description="boneIO app for BeagleBone Black.",
    )
    parser.add_argument(ACTION, type=str, default="run")
    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument(
        "-c",
        "--config",
        metavar="path_to_config_dir",
        default="./config.yaml",
        help="File which contains boneIO configuration",
    )
    parser.add_argument(
        "--debug", "-d", action="count", help="Start boneIO in debug mode", default=0
    )
    parser.add_argument(
        "--mqttusername", help="Mqtt username to use if you don't want provide in file."
    )
    parser.add_argument(
        "--mqttpassword", help="Mqtt password to use if you don't want provide in file."
    )
    arguments = parser.parse_args()

    return arguments


def run(config: str, debug: int, mqttusername: str = "", mqttpassword: str = ""):
    """Run BoneIO."""
    _LOGGER.info("BoneIO %s starting.", __version__)
    try:
        _config = load_config_from_file(config_file=config)
        if not _config:
            _LOGGER.error("Config not loaded. Exiting.")
            return 1
        configure_logger(log_config=_config.get("logger"), debug=debug)
        return asyncio.run(
            async_run(
                config=_config,
                config_file=config,
                mqttusername=mqttusername,
                mqttpassword=mqttpassword,
            ),
        )
    except (ConfigurationException, MarkedYAMLError) as err:
        _LOGGER.error("Failed to load config. %s Exiting.", err)
        return 1


def main() -> int:
    """Start boneIO."""

    args = get_arguments()
    debug = args.debug

    exit_code = 0
    if args.action == "run":
        exit_code = run(
            config=args.config,
            mqttusername=args.mqttusername,
            mqttpassword=args.mqttpassword,
            debug=debug,
        )

    return exit_code


if __name__ == "__main__":
    sys.exit(main())



import asyncio
import logging

from boneio.helper import load_config_from_file
from boneio.manager import Manager

_LOGGER = logging.getLogger(__name__)

from boneio.const import LM75, MCP23017, MCP_TEMP_9808, OUTPUT


async def test_32_5():
    _config = load_config_from_file("./32_5_config.yaml")
    if not _config:
        _LOGGER.info("Missing file.")
        return
    manager = Manager(
        send_message=lambda topic, payload: None,
        relay_pins=_config.get(OUTPUT, []),
        sensors={
            LM75: _config.get(LM75),
            MCP_TEMP_9808: _config.get(MCP_TEMP_9808),
        },
        mcp23017=_config.get(MCP23017, []),
        ha_discovery=False,
    )
    await asyncio.sleep(1)
    for key, value in manager.output.items():
        print(
            f"State of \33[42m{key} MCP {value.mcp_id} pin: {value.pin_id}\033[0m is {value.is_active}. Toggle it."
        )
        value.toggle()
        print("It should be on. Sleeping for 2 secs.")
        await asyncio.sleep(2)


if __name__ == "__main__":
    asyncio.run(test_32_5())



"""DS2482 base class.
Module by https://github.com/fgervais/ds2482
"""
import time

import adafruit_bus_device.i2c_device as i2c_device

# Default i2c address (AD0 = GND, AD1 = GND)
DS2482_ADDRESS = 0x18

# DS2482 device commands
COMMAND_DEVICE_RESET = 0xF0
COMMAND_SET_POINTER = 0xE1
COMMAND_WRITE_CONFIG = 0xD2
COMMAND_1W_RESET = 0xB4
COMMAND_1W_SINGLE_BIT = 0x87
COMMAND_1W_WRITE_BYTE = 0xA5
COMMAND_1W_READ_BYTE = 0x96
COMMAND_1W_TRIPLET = 0x78

# DS2482 read pointer codes
POINTER_STATUS = 0xF0
POINTER_DATA = 0xE1
POINTER_CONFIG = 0xC3

# DS2482 configuration register
CONFIG_ACTIVE_PULLUP = 0x01
CONFIG_STRONG_PULLUP = 0x04
CONFIG_1W_OVERDRIVE = 0x08

CONFIG_MASK = 0x0F

# DS2482 status register
STATUS_1W_BUSY = 0x01
STATUS_PRESENCE_PULSE = 0x02
STATUT_SHORT_DETECTED = 0x04
STATUS_LOGIC_LEVEL = 0x08
STATUS_DEVICE_RESET = 0x10
STATUS_SINGLE_BIT = 0x20
STATUS_TRIPLET_BIT = 0x40
STATUS_BRANCH_TAKEN = 0x80


class DS2482:
    def __init__(self, i2c, address=DS2482_ADDRESS, active_pullup=False):
        self._i2c = i2c_device.I2CDevice(i2c, address)
        self.device_reset()
        if active_pullup:
            self.device_config = CONFIG_ACTIVE_PULLUP

        # 1-Wire bus busy with STRONG_PULLUP
        self._bus_busy = time.monotonic()

    def device_reset(self):
        """Terminate any 1-wire communication and reset the DS2482"""
        with self._i2c as i2c:
            i2c.write(bytes([COMMAND_DEVICE_RESET]))

    @property
    def device_status(self):
        with self._i2c as i2c:
            buf = bytearray([COMMAND_SET_POINTER, POINTER_STATUS])
            i2c.write_then_readinto(buf, buf, in_end=1)
            return buf[0]

    @property
    def device_config(self):
        with self._i2c as i2c:
            buf = bytearray([COMMAND_SET_POINTER, POINTER_CONFIG])
            i2c.write_then_readinto(buf, buf, in_end=1)
            return buf[0] & 0x0F

    @device_config.setter
    def device_config(self, config):
        with self._i2c as i2c:
            i2c.write(
                bytes([COMMAND_WRITE_CONFIG, (config & 0x0F) | ((~config << 4) & 0xF0)])
            )

    def reset(self):
        with self._i2c as i2c:
            buf = bytearray([COMMAND_1W_RESET])
            i2c.write(buf)
            while True:
                i2c.readinto(buf)
                if not buf[0] & STATUS_1W_BUSY:
                    break
                time.sleep(0.001)
            return not buf[0] & STATUS_PRESENCE_PULSE

    def single_bit(self, bit=1, strong_pullup=False, busy=None):
        with self._i2c as i2c:
            buf = bytearray(2)
            if strong_pullup:
                buf[0:2] = [COMMAND_SET_POINTER, POINTER_CONFIG]
                i2c.write_then_readinto(buf, buf, in_start=1)
                buf[0] = COMMAND_WRITE_CONFIG
                buf[1] = buf[1] & CONFIG_MASK | CONFIG_STRONG_PULLUP
                buf[1] |= (~buf[1] << 4) & 0xFF
                i2c.write(buf)
            buf[0:2] = [COMMAND_1W_SINGLE_BIT, 0x80 if bit else 0x00]
            i2c.write(buf)
            while True:
                i2c.readinto(buf, end=1)
                if not buf[0] & STATUS_1W_BUSY:
                    break
                time.sleep(0.001)
            if busy:
                self._bus_busy = time.monotonic() + busy
            return buf[0] & STATUS_SINGLE_BIT != 0

    def write_byte(self, data, strong_pullup=False, busy=None):
        with self._i2c as i2c:
            buf = bytearray(2)
            if strong_pullup:
                buf[0:2] = [COMMAND_SET_POINTER, POINTER_CONFIG]
                i2c.write_then_readinto(buf, buf, in_start=1)
                buf[0] = COMMAND_WRITE_CONFIG
                buf[1] = buf[1] & CONFIG_MASK | CONFIG_STRONG_PULLUP
                buf[1] |= (~buf[1] << 4) & 0xFF
                i2c.write(buf)
            buf[0:2] = [COMMAND_1W_WRITE_BYTE, data]
            i2c.write(buf)
            while True:
                i2c.readinto(buf, end=1)
                if buf[0] & STATUS_1W_BUSY == 0:
                    break
                time.sleep(0.001)
            if busy:
                self._bus_busy = time.monotonic() + busy

    def read_byte(self):
        with self._i2c as i2c:
            buf = bytearray([COMMAND_1W_READ_BYTE, 0x00])
            i2c.write(buf, end=1)
            while True:
                i2c.readinto(buf, end=1)
                if not buf[0] & STATUS_1W_BUSY:
                    break
                time.sleep(0.001)
            buf[0:2] = [COMMAND_SET_POINTER, POINTER_DATA]
            i2c.write_then_readinto(buf, buf, in_end=1)
            return buf[0]

    def wait_ready(self):
        while True:
            t = self._bus_busy - time.monotonic()
            if t > 0:
                time.sleep(t)
            else:
                break
        with self._i2c as i2c:
            buf = bytearray([COMMAND_SET_POINTER, POINTER_STATUS])
            i2c.write(buf)
            while True:
                i2c.readinto(buf, end=1)
                if not buf[0] & STATUS_1W_BUSY:
                    break
                time.sleep(0.001)
            return buf[0]

    def triplet(self, dir):
        pass



import asyncio
import json
import logging
import os
from datetime import datetime
from struct import unpack

from boneio.const import (
    ADDRESS,
    BASE,
    LENGTH,
    MODEL,
    OFFLINE,
    ONLINE,
    REGISTERS,
    SENSOR,
    STATE,
)
from boneio.helper import BasicMqtt, AsyncUpdater
from boneio.helper.config import ConfigHelper
from boneio.helper.events import EventBus
from boneio.helper.ha_discovery import modbus_sensor_availabilty_message
from boneio.modbus import Modbus

_LOGGER = logging.getLogger(__name__)


def float32(result, base, addr):
    """Read Float value from register."""
    low = result.getRegister(addr - base)
    high = result.getRegister(addr - base + 1)
    data = bytearray(4)
    data[0] = high & 0xFF
    data[1] = high >> 8
    data[2] = low & 0xFF
    data[3] = low >> 8
    val = unpack("f", bytes(data))
    return val[0]


def floatsofar(result, base, addr):
    """Read Float value from register."""
    low = result.getRegister(addr - base)
    high = result.getRegister(addr - base + 1)
    return high + low


def multiply0_1(result, base, addr):
    low = result.getRegister(addr - base)
    return round(low * 0.1, 4)


def multiply0_01(result, base, addr):
    low = result.getRegister(addr - base)
    return round(low * 0.01, 4)


def multiply10(result, base, addr):
    low = result.getRegister(addr - base)
    return round(low * 10, 4)


def regular_result(result, base, addr):
    return result.getRegister(addr - base)


CONVERT_METHODS = {
    "float32": float32,
    "multiply0_1": multiply0_1,
    "multiply0_01": multiply0_01,
    "floatsofar": floatsofar,
    "multiply10": multiply10,
    "regular": regular_result,
}
REGISTERS_BASE = "registers_base"


def open_json(model: str) -> dict:
    """Open json file."""
    file = f"{os.path.join(os.path.dirname(__file__))}/{model}.json"
    with open(file, "r") as db_file:
        datastore = json.load(db_file)
        return datastore


class ModbusSensor(BasicMqtt, AsyncUpdater):
    """Represent Modbus sensor in BoneIO."""

    SensorClass = None
    DefaultName = "ModbusSensor"

    def __init__(
        self,
        modbus: Modbus,
        address: str,
        model: str,
        config_helper: ConfigHelper,
        event_bus: EventBus,
        id: str = DefaultName,
        **kwargs,
    ):
        """Initialize Modbus sensor class."""
        super().__init__(
            id=id or address,
            topic_type=SENSOR,
            topic_prefix=config_helper.topic_prefix,
            **kwargs,
        )
        self._config_helper = config_helper
        self._modbus = modbus
        self._db = open_json(model=model)
        self._model = self._db[MODEL]
        self._address = address
        self._discovery_sent = False
        self._payload_online = OFFLINE
        event_bus.add_haonline_listener(target=self.set_payload_offline)
        AsyncUpdater.__init__(self, **kwargs)

    def set_payload_offline(self):
        self._payload_online = OFFLINE

    def _send_ha_autodiscovery(
        self, id: str, sdm_name: str, sensor_id: str, **kwargs
    ) -> None:
        """Send HA autodiscovery information for each Modbus sensor."""
        _LOGGER.debug(
            "Sending HA discovery for modbus sensor %s %s.", sdm_name, sensor_id
        )
        topic = (
            f"{self._config_helper.ha_discovery_prefix}/{SENSOR}/{self._config_helper.topic_prefix}{id}"
            f"/{id}{sensor_id.replace('_', '').replace(' ', '').lower()}/config"
        )
        payload = modbus_sensor_availabilty_message(
            topic=self._config_helper.topic_prefix,
            id=id,
            name=sdm_name,
            model=self._model,
            sensor_id=sensor_id,
            **kwargs,
        )
        self._config_helper.add_autodiscovery_msg(topic=topic, payload=payload)
        self._send_message(topic=topic, payload=payload)

    def _send_discovery_for_all_registers(self, register: int = 0) -> bool:
        """Send discovery message to HA for each register."""
        for data in self._db[REGISTERS_BASE]:
            for register in data[REGISTERS]:
                value_template = (
                    f'{{{{ value_json.{register.get("name").replace(" ", "")} | '
                    f'{register.get("ha_filter", "round(2)")} }}}}'
                )
                kwargs = {
                    "unit_of_measurement": register.get("unit_of_measurement"),
                    "state_class": register.get("state_class"),
                    "value_template": value_template,
                    "sensor_id": register.get("name"),
                }
                device_class = register.get("device_class")
                if device_class:
                    kwargs["device_class"] = device_class
                self._send_ha_autodiscovery(
                    id=self._id,
                    sdm_name=self._name,
                    state_topic_base=data[BASE],
                    **kwargs,
                )
        return datetime.now()

    async def check_availability(self) -> None:
        """Get first register and check if it's available."""
        if (
            not self._discovery_sent
            or (datetime.now() - self._discovery_sent).seconds > 3600
        ) and self._config_helper.topic_prefix:
            self._discovery_sent = False
            first_register_base = self._db[REGISTERS_BASE][0]
            register_method = first_register_base.get("register_type", "input")
            # Let's try fetch register 2 times in case something wrong with initial packet.
            for _ in [0, 1]:
                register = await self._modbus.read_single_register(
                    unit=self._address,
                    address=first_register_base[REGISTERS][0][ADDRESS],
                    method=register_method,
                )
                if register is not None:
                    self._discovery_sent = self._send_discovery_for_all_registers(
                        register
                    )
                    await asyncio.sleep(2)
                    break
            _LOGGER.error(
                "Discovery for %s not sent. First register not available.", self._id
            )

    async def async_update(self, time: datetime) -> None:
        """Fetch state periodically and send to MQTT."""
        update_interval = self._update_interval.total_in_seconds
        await self.check_availability()
        for data in self._db[REGISTERS_BASE]:
            values = await self._modbus.read_multiple_registers(
                unit=self._address,
                address=data[BASE],
                count=data[LENGTH],
                method=data.get("register_type", "input"),
            )
            if self._payload_online == OFFLINE and values:
                _LOGGER.info("Sending online payload about device.")
                self._payload_online = ONLINE
                self._send_message(
                    topic=f"{self._config_helper.topic_prefix}/{self._id}{STATE}",
                    payload=self._payload_online,
                )
            if not values:
                if update_interval < 600:
                    # Let's wait litte more for device.
                    update_interval = update_interval * 1.5
                else:
                    # Let's assume device is offline.
                    self.set_payload_offline()
                    self._send_message(
                        topic=f"{self._config_helper.topic_prefix}/{self._id}{STATE}",
                        payload=self._payload_online,
                    )
                _LOGGER.warn(
                    "Can't fetch data from modbus device %s. Will sleep for %s seconds",
                    self.id,
                    update_interval,
                )
                self._discovery_sent = False
                return update_interval
            elif update_interval != self._update_interval.total_in_seconds:
                update_interval = self._update_interval.total_in_seconds
            output = {}
            for register in data["registers"]:
                output[register.get("name").replace(" ", "")] = CONVERT_METHODS[
                    register.get("return_type", "regular")
                ](result=values, base=data[BASE], addr=register.get("address"))
            self._send_message(
                topic=f"{self._send_topic}/{data[BASE]}",
                payload=output,
            )
        return update_interval



from boneio.const import (
    CLOSE,
    CLOSED,
    CLOSING,
    COVER,
    INPUT,
    INPUT_SENSOR,
    OFF,
    ON,
    OPEN,
    OPENING,
    RELAY,
    SENSOR,
    STATE,
    STOP,
)
from boneio.version import __version__


def ha_availabilty_message(
    id: str,
    name: str,
    topic: str = "boneIO",
    device_type: str = INPUT,
    model: str = "boneIO Relay Board",
    **kwargs,
):
    """Create availability topic for HA."""
    return {
        "availability": [{"topic": f"{topic}/{STATE}"}],
        "device": {
            "identifiers": [topic],
            "manufacturer": "boneIO",
            "model": model,
            "name": f"boneIO {topic}",
            "sw_version": __version__,
        },
        "name": name,
        "state_topic": f"{topic}/{device_type}/{id}",
        "unique_id": f"{topic}{device_type}{id}",
        **kwargs,
    }


def ha_light_availabilty_message(id: str, topic: str = "boneIO", **kwargs):
    """Create LIGHT availability topic for HA."""
    msg = ha_availabilty_message(device_type=RELAY, topic=topic, id=id, **kwargs)
    msg["command_topic"] = f"{topic}/cmd/relay/{id}/set"
    msg["payload_off"] = OFF
    msg["payload_on"] = ON
    msg["state_value_template"] = "{{ value_json.state }}"
    return msg


def ha_button_availabilty_message(id: str, topic: str = "boneIO", **kwargs):
    """Create BUTTON availability topic for HA."""
    msg = ha_availabilty_message(device_type="button", topic=topic, id=id, **kwargs)
    msg["command_topic"] = f"{topic}/cmd/button/{id}/set"
    msg["payload_press"] = "reload"
    return msg


def ha_switch_availabilty_message(id: str, topic: str = "boneIO", **kwargs):
    """Create SWITCH availability topic for HA."""
    msg = ha_availabilty_message(device_type=RELAY, topic=topic, id=id, **kwargs)
    msg["command_topic"] = f"{topic}/cmd/relay/{id}/set"
    msg["payload_off"] = OFF
    msg["payload_on"] = ON
    msg["value_template"] = "{{ value_json.state }}"
    return msg


def ha_input_availabilty_message(**kwargs):
    msg = ha_availabilty_message(device_type=INPUT, **kwargs)
    msg["icon"] = "mdi:gesture-double-tap"
    return msg


def ha_adc_sensor_availabilty_message(**kwargs):
    msg = ha_availabilty_message(device_type=SENSOR, **kwargs)
    msg["unit_of_measurement"] = "V"
    msg["device_class"] = "voltage"
    msg["state_class"] = "measurement"
    return msg


def ha_sensor_availabilty_message(unit_of_measurement: str = None, **kwargs):
    msg = ha_availabilty_message(device_type=SENSOR, **kwargs)
    if not unit_of_measurement:
        return msg


def ha_binary_sensor_availabilty_message(id: str, name: str, topic: str = "boneIO"):
    """Create availability topic for HA."""
    return {
        "availability": [{"topic": f"{topic}/{STATE}"}],
        "device": {
            "identifiers": [topic],
            "manufacturer": "boneIO",
            "model": "boneIO Relay Board",
            "name": f"boneIO {topic}",
            "sw_version": __version__,
        },
        "payload_on": "pressed",
        "payload_off": "released",
        "name": name,
        "state_topic": f"{topic}/{INPUT_SENSOR}/{id}",
        "unique_id": f"{topic}{INPUT_SENSOR}{id}",
    }


def ha_sensor_temp_availabilty_message(id: str, name: str, topic: str = "boneIO"):
    """Create availability topic for HA."""
    return {
        "availability": [{"topic": f"{topic}/{STATE}"}],
        "device": {
            "identifiers": [topic],
            "manufacturer": "boneIO",
            "model": "boneIO Relay Board",
            "name": f"boneIO {topic}",
            "sw_version": __version__,
        },
        "name": name,
        "state_topic": f"{topic}/{SENSOR}/{id}",
        "unique_id": f"{topic}{SENSOR}{id}",
        "device_class": "temperature",
        "state_class": "measurement",
        "unit_of_measurement": "°C",
        "value_template": "{{ value_json.state }}",
    }


def modbus_sensor_availabilty_message(
    id: str,
    sensor_id: str,
    name: str,
    state_topic_base: str,
    topic: str,
    model: str,
    device_type: str = SENSOR,
    **kwargs,
):
    """Create Modbus Sensor availability topic for HA."""
    return {
        "availability": [{"topic": f"{topic}/{id}{STATE}"}],
        "device": {
            "identifiers": [id],
            "manufacturer": "boneIO",
            "model": model,
            "name": f"boneIO {name.upper()}",
            "sw_version": __version__,
        },
        "name": sensor_id,
        "state_topic": f"{topic}/{device_type}/{id}/{state_topic_base}",
        "unique_id": f"{topic}{sensor_id.replace('_', '').lower()}{name.lower()}",
        **kwargs,
    }


def ha_cover_availabilty_message(
    id: str, name: str, device_class: str, topic: str = "boneIO"
):
    """Create Cover availability topic for HA."""
    kwargs = {"device_class": device_class} if device_class else {}
    msg = ha_availabilty_message(
        device_type=COVER, topic=topic, id=id, name=name, **kwargs
    )

    return {
        **msg,
        "command_topic": f"{topic}/cmd/cover/{id}/set",
        "set_position_topic": f"{topic}/cmd/cover/{id}/pos",
        "payload_open": OPEN,
        "payload_close": CLOSE,
        "payload_stop": STOP,
        "state_open": OPEN,
        "state_opening": OPENING,
        "state_closed": CLOSED,
        "state_closing": CLOSING,
        "state_topic": f"{topic}/{COVER}/{id}/state",
        "position_topic": f"{topic}/{COVER}/{id}/pos",
    }



"""
Special unique implementation of asyncio.Queue for boneIO.
If MQTT is down then regular queue can append multiple ON/OFF for same topic.
After re-connection it would send all messages. It's not necessary, last payload of same topic is enough.
"""
import asyncio


class UniqueQueue(asyncio.Queue):
    """Unique implementation of asyncio.Queue."""

    def _init(self, maxsize):
        """Create unique dict of tuple[0]."""
        super()._init(maxsize=maxsize)
        self._unique_set = {}

    def _put(self, item):
        """If item does not exists add it.
        If exists remove old one and put new one in the beginning of the queue."""
        if item[0] not in self._unique_set:
            super()._put(item)
            self._unique_set[item[0]] = item
        else:
            self._queue.remove(self._unique_set[item[0]])
            super()._put(item)
            self._unique_set[item[0]] = item

    def _get(self):
        """Get item and remove it."""
        item = super()._get()
        del self._unique_set[item[0]]
        return item



"""GPIO Relay module.
!!Not used in BoneIO.
Created just in case.
"""

import logging

from boneio.const import HIGH, LOW
from boneio.helper import read_input, setup_output, write_output
from boneio.relay.basic import BasicRelay

_LOGGER = logging.getLogger(__name__)


class GpioRelay(BasicRelay):
    """Represents GPIO Relay output"""

    def __init__(self, pin: str, **kwargs) -> None:
        """Initialize Gpio relay."""
        super().__init(**kwargs)
        self._pin = pin
        setup_output(self._pin)
        write_output(self.pin, LOW)
        _LOGGER.debug("Setup relay with pin %s", self._pin)

    @property
    def is_active(self) -> bool:
        """Is relay active."""
        return read_input(self.pin, on_state=HIGH)

    @property
    def pin(self) -> str:
        """PIN of the relay"""
        return self._pin

    def turn_on(self) -> None:
        """Call turn on action."""
        write_output(self.pin, HIGH)
        self._loop.call_soon_threadsafe(self.send_state)

    def turn_off(self) -> None:
        """Call turn off action."""
        write_output(self.pin, LOW)
        self._loop.call_soon_threadsafe(self.send_state)



from typing import Any, Callable, TypeVar

CALLABLE_T = TypeVar("CALLABLE_T", bound=Callable[..., Any])
CALLBACK_TYPE = Callable[[], None]


def callback(func: CALLABLE_T) -> CALLABLE_T:
    """Annotation to mark method as safe to call from within the event loop."""
    setattr(func, "_boneio_callback", True)
    return func


def is_callback(func: Callable[..., Any]) -> bool:
    """Check if function is safe to be called in the event loop."""
    return getattr(func, "_boneio_callback", False) is True



"""GPIOInputButton to receive signals."""
import logging
from datetime import datetime, timedelta
from functools import partial

from boneio.const import DOUBLE, LONG, SINGLE
from boneio.helper import GpioBaseClass, edge_detect

# TIMINGS FOR BUTTONS
DEBOUNCE_DURATION = timedelta(microseconds=150000)
LONG_PRESS_DURATION = timedelta(microseconds=700000)
DELAY_DURATION = 0.08
SECOND_DELAY_DURATION = 0.14

_LOGGER = logging.getLogger(__name__)


class GpioInputButton(GpioBaseClass):
    """Represent Gpio input switch."""

    def __init__(self, **kwargs) -> None:
        """Setup GPIO Input Button"""
        super().__init__(**kwargs)
        edge_detect(
            self._pin,
            callback=self._handle_press,
            bounce=self._bounce_time.total_milliseconds,
        )
        self._first_press_timestamp = None
        self._is_long_press = False
        self._second_press_timestamp = None
        self._second_check = False
        _LOGGER.debug("Configured listening for input pin %s", self._pin)

    def _handle_press(self, pin: str) -> None:
        """Handle the button press callback"""
        # Ignore if we are in a long press
        if self._is_long_press:
            return
        now = datetime.now()

        # Debounce button
        if (
            self._first_press_timestamp is not None
            and now - self._first_press_timestamp < DEBOUNCE_DURATION
        ):
            return

        # Second click debounce. Just in case.
        if (
            self._second_press_timestamp is not None
            and now - self._second_press_timestamp < DEBOUNCE_DURATION
        ):
            return
        if not self._first_press_timestamp:
            self._first_press_timestamp = now
        elif not self._second_press_timestamp:
            self._second_press_timestamp = now

        self._loop.call_soon_threadsafe(
            self._loop.call_later,
            DELAY_DURATION,
            self.check_press_length,
        )

    def check_press_length(self) -> None:
        """Check if it's a single, double or long press"""
        # Check if button is still pressed
        if self.is_pressed:
            # Schedule a new check
            self._loop.call_soon_threadsafe(
                self._loop.call_later,
                DELAY_DURATION,
                self.check_press_length,
            )

            # Handle edge case due to multiple clicks
            if self._first_press_timestamp is None:
                return

            # Check if we reached a long press
            diff = datetime.now() - self._first_press_timestamp
            if not self._is_long_press and diff > LONG_PRESS_DURATION:
                self._is_long_press = True
                _LOGGER.debug("Long button press on pin %s, call callback", self._pin)
                self._loop.call_soon_threadsafe(
                    partial(self._press_callback, LONG, self._pin)
                )
            return

        # Handle short press
        if not self._is_long_press:
            if not self._second_press_timestamp and not self._second_check:
                # let's try to check if second click will atempt
                self._second_check = True
                self._loop.call_soon_threadsafe(
                    self._loop.call_later,
                    SECOND_DELAY_DURATION,
                    self.check_press_length,
                )
                return
            if self._second_check:
                if self._second_press_timestamp:
                    _LOGGER.debug(
                        "Double click event on pin %s, diff %s",
                        self._pin,
                        self._second_press_timestamp - self._first_press_timestamp,
                    )
                    self._loop.call_soon_threadsafe(
                        partial(self._press_callback, DOUBLE, self._pin)
                    )

                elif self._first_press_timestamp:
                    _LOGGER.debug("One click event on pin %s, call callback", self._pin)
                    self._loop.call_soon_threadsafe(
                        partial(self._press_callback, SINGLE, self._pin)
                    )

        # Clean state on button released
        self._first_press_timestamp = None
        self._second_press_timestamp = None
        self._second_check = False
        self._is_long_press = False



"""MCP23017 Relay module."""

import logging

from adafruit_mcp230xx.mcp23017 import MCP23017, DigitalInOut

from boneio.const import NONE, SWITCH
from boneio.helper.events import async_track_point_in_time, utcnow
from boneio.helper.util import callback
from boneio.relay.basic import BasicRelay

_LOGGER = logging.getLogger(__name__)


class MCPRelay(BasicRelay):
    """Represents MCP Relay output"""

    def __init__(
        self,
        pin: int,
        mcp: MCP23017,
        mcp_id: str,
        output_type: str = SWITCH,
        restored_state: bool = False,
        **kwargs
    ) -> None:
        """Initialize MCP relay."""
        self._pin: DigitalInOut = mcp.get_pin(pin)
        if output_type == NONE:
            """Just in case to not restore state of covers etc."""
            restored_state = False
        self._pin.switch_to_output(value=restored_state)
        super().__init__(
            **kwargs, output_type=output_type, restored_state=restored_state
        )
        self._pin_id = pin
        self._mcp_id = mcp_id
        _LOGGER.debug("Setup MCP with pin %s", self._pin_id)

    @property
    def is_mcp_type(self) -> bool:
        """Check if relay is mcp type."""
        return True

    @property
    def pin_id(self) -> str:
        """Return PIN id."""
        return self._pin_id

    @property
    def mcp_id(self) -> str:
        """Retrieve parent MCP ID."""
        return self._mcp_id

    @property
    def is_active(self) -> bool:
        """Is relay active."""
        return self.pin.value

    @property
    def pin(self) -> str:
        """PIN of the relay"""
        return self._pin

    @callback
    def _momentary_callback(self, time, action):
        _LOGGER.info("Momentary callback at %s", time)
        action()

    def turn_on(self) -> None:
        """Call turn on action."""
        self.pin.value = True
        if self._momentary_turn_on:
            async_track_point_in_time(
                loop=self._loop,
                action=lambda x: self._momentary_callback(time=x, action=self.turn_off),
                point_in_time=utcnow() + self._momentary_turn_on.as_timedelta,
            )
        self._loop.call_soon_threadsafe(self.send_state)

    def turn_off(self) -> None:
        """Call turn off action."""
        self.pin.value = False
        if self._momentary_turn_off:
            async_track_point_in_time(
                loop=self._loop,
                action=lambda x: self._momentary_callback(time=x, action=self.turn_on),
                point_in_time=utcnow() + self._momentary_turn_off.as_timedelta,
            )
        self._loop.call_soon_threadsafe(self.send_state)



from __future__ import annotations

import asyncio
import logging
from typing import Any

from pymodbus.client.sync import BaseModbusClient, ModbusSerialClient
from pymodbus.constants import Endian
from pymodbus.exceptions import ModbusException
from pymodbus.payload import BinaryPayloadDecoder
from pymodbus.pdu import ModbusResponse
from pymodbus.register_read_message import ReadInputRegistersResponse

from boneio.const import ID, REGISTERS, RX, TX, UART
from boneio.helper import configure_pin
from boneio.helper.exceptions import ModbusUartException

_LOGGER = logging.getLogger(__name__)


class Modbus:
    """Represent modbus connection over chosen UART."""

    def __init__(self, uart: dict[str, Any]) -> None:
        """Initialize the Modbus hub."""
        rx = uart.get(RX)
        tx = uart.get(TX)
        if not tx or not rx:
            raise ModbusUartException

        configure_pin(pin=rx, mode=UART)
        configure_pin(pin=tx, mode=UART)
        self._uart = uart

        # generic configuration
        self._client: BaseModbusClient | None = None
        self._lock = asyncio.Lock()
        try:
            self._client = ModbusSerialClient(
                port=self._uart[ID],
                method="rtu",
                baudrate=9600,
                stopbits=1,
                bytesize=8,
                parity="N",
            )
            self._read_methods = {
                "input": self._client.read_input_registers,
                "holding": self._client.read_holding_registers,
            }
        except ModbusException as exception_error:
            _LOGGER.error(exception_error)

    async def async_close(self) -> None:
        """Disconnect client."""
        async with self._lock:
            if self._client:
                try:
                    self._client.close()
                except ModbusException as exception_error:
                    _LOGGER.error(exception_error)
                del self._client
                self._client = None
                _LOGGER.warning("modbus communication closed")

    def _pymodbus_connect(self) -> bool:
        """Connect client."""
        try:
            return self._client.connect()  # type: ignore[union-attr]
        except ModbusException as exception_error:
            _LOGGER.error(exception_error)
            return False

    async def read_single_register(
        self, unit: int, address: int, count: int = 2, method: str = "input"
    ) -> float | None:
        """Call sync. pymodbus."""
        async with self._lock:
            if not self._pymodbus_connect():
                _LOGGER.error("Can't connect to Modbus address %s.", address)
                return None
            kwargs = {"unit": unit, "count": count} if unit else {}
            try:
                result: ReadInputRegistersResponse = self._read_methods[method](
                    address, **kwargs
                )
            except ModbusException as exception_error:
                _LOGGER.error(exception_error)
                return None
            if not hasattr(result, REGISTERS):
                _LOGGER.error(str(result))
                return None
            return BinaryPayloadDecoder.fromRegisters(
                result.registers, byteorder=Endian.Big, wordorder=Endian.Big
            ).decode_32bit_float()

    async def read_multiple_registers(
        self, unit: int, address: int, count: int = 2, method: str = "input"
    ) -> ModbusResponse:
        """Call sync. pymodbus."""
        async with self._lock:
            if not self._pymodbus_connect():
                _LOGGER.error("Can't connect to Modbus.")
                return None
            kwargs = {"unit": unit, "count": count} if unit else {}
            try:
                result: ReadInputRegistersResponse = self._read_methods[method](
                    address, **kwargs
                )
            except ModbusException as exception_error:
                _LOGGER.error(exception_error)
                return None
            if not hasattr(result, REGISTERS):
                _LOGGER.error(str(result))
                return None
            return result



import math
from collections import OrderedDict
from datetime import timedelta


def is_approximately_integer(value):
    if isinstance(value, int):
        return True
    return abs(value - round(value)) < 0.001


class TimePeriod:
    def __init__(
        self,
        microseconds=None,
        milliseconds=None,
        seconds=None,
        minutes=None,
        hours=None,
        days=None,
    ) -> None:
        if days is not None:
            if not is_approximately_integer(days):
                frac_days, days = math.modf(days)
                hours = (hours or 0) + frac_days * 24
            self.days = int(round(days))
        else:
            self.days = None

        if hours is not None:
            if not is_approximately_integer(hours):
                frac_hours, hours = math.modf(hours)
                minutes = (minutes or 0) + frac_hours * 60
            self.hours = int(round(hours))
        else:
            self.hours = None

        if minutes is not None:
            if not is_approximately_integer(minutes):
                frac_minutes, minutes = math.modf(minutes)
                seconds = (seconds or 0) + frac_minutes * 60
            self.minutes = int(round(minutes))
        else:
            self.minutes = None

        if seconds is not None:
            if not is_approximately_integer(seconds):
                frac_seconds, seconds = math.modf(seconds)
                milliseconds = (milliseconds or 0) + frac_seconds * 1000
            self.seconds = int(round(seconds))
        else:
            self.seconds = None

        if milliseconds is not None:
            if not is_approximately_integer(milliseconds):
                frac_milliseconds, milliseconds = math.modf(milliseconds)
                microseconds = (microseconds or 0) + frac_milliseconds * 1000
            self.milliseconds = int(round(milliseconds))
        else:
            self.milliseconds = None

        if microseconds is not None:
            if not is_approximately_integer(microseconds):
                raise ValueError("Maximum precision is microseconds")
            self.microseconds = int(round(microseconds))
        else:
            self.microseconds = None

        self._timedelta = timedelta(
            days=self.days or 0,
            hours=self.hours or 0,
            minutes=self.minutes or 0,
            seconds=self.seconds or 0,
            milliseconds=self.milliseconds or 0,
            microseconds=self.microseconds or 0,
        )
        self._total_in_seconds = self.total_microseconds / 1000000.0

    @property
    def as_timedelta(self) -> timedelta:
        return self._timedelta

    def as_dict(self) -> OrderedDict:
        out = OrderedDict()
        if self.microseconds is not None:
            out["microseconds"] = self.microseconds
        if self.milliseconds is not None:
            out["milliseconds"] = self.milliseconds
        if self.seconds is not None:
            out["seconds"] = self.seconds
        if self.minutes is not None:
            out["minutes"] = self.minutes
        if self.hours is not None:
            out["hours"] = self.hours
        if self.days is not None:
            out["days"] = self.days
        return out

    def __str__(self) -> str:
        if self.microseconds is not None:
            return f"{self.total_microseconds}us"
        if self.milliseconds is not None:
            return f"{self.total_milliseconds}ms"
        if self.seconds is not None:
            return f"{self.total_seconds}s"
        if self.minutes is not None:
            return f"{self.total_minutes}min"
        if self.hours is not None:
            return f"{self.total_hours}h"
        if self.days is not None:
            return f"{self.total_days}d"
        return "0s"

    def __repr__(self):
        return f"TimePeriod<{self.__str__()}>"

    @property
    def total_in_seconds(self) -> float:
        return self._total_in_seconds

    @property
    def total_microseconds(self):
        return self.total_milliseconds * 1000 + (self.microseconds or 0)

    @property
    def total_milliseconds(self):
        return self.total_seconds * 1000 + (self.milliseconds or 0)

    @property
    def total_seconds(self):
        return self.total_minutes * 60 + (self.seconds or 0)

    @property
    def total_minutes(self):
        return self.total_hours * 60 + (self.minutes or 0)

    @property
    def total_hours(self):
        return self.total_days * 24 + (self.hours or 0)

    @property
    def total_days(self):
        return self.days or 0

    def __eq__(self, other):
        if isinstance(other, TimePeriod):
            return self.total_microseconds == other.total_microseconds
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, TimePeriod):
            return self.total_microseconds != other.total_microseconds
        return NotImplemented

    def __lt__(self, other):
        if isinstance(other, TimePeriod):
            return self.total_microseconds < other.total_microseconds
        return NotImplemented

    def __gt__(self, other):
        if isinstance(other, TimePeriod):
            return self.total_microseconds > other.total_microseconds
        return NotImplemented

    def __le__(self, other):
        if isinstance(other, TimePeriod):
            return self.total_microseconds <= other.total_microseconds
        return NotImplemented

    def __ge__(self, other):
        if isinstance(other, TimePeriod):
            return self.total_microseconds >= other.total_microseconds
        return NotImplemented


class TimePeriodMicroseconds(TimePeriod):
    pass


class TimePeriodMilliseconds(TimePeriod):
    pass


class TimePeriodSeconds(TimePeriod):
    pass


class TimePeriodMinutes(TimePeriod):
    pass



"""
Module to provide basic config options.
"""
from _collections_abc import dict_values

from boneio.const import BONEIO, HOMEASSISTANT


class ConfigHelper:
    def __init__(
        self,
        topic_prefix: str = BONEIO,
        ha_discovery: bool = True,
        ha_discovery_prefix: str = HOMEASSISTANT,
    ):
        self._topic_prefix = topic_prefix
        self._ha_discovery = ha_discovery
        self._ha_discovery_prefix = ha_discovery_prefix
        self._autodiscovery_messages = {}

    @property
    def topic_prefix(self) -> str:
        return self._topic_prefix

    @property
    def ha_discovery(self) -> bool:
        return self._ha_discovery

    @property
    def ha_discovery_prefix(self) -> str:
        return self._ha_discovery_prefix

    @property
    def cmd_topic_prefix(self) -> str:
        return f"{self.topic_prefix}/cmd/"

    @property
    def subscribe_topic(self) -> str:
        return f"{self.cmd_topic_prefix}+/+/#"

    def add_autodiscovery_msg(self, topic: str, payload: str):
        """Add autodiscovery message."""
        self._autodiscovery_messages[topic] = {"topic": topic, "payload": payload}

    @property
    def autodiscovery_msgs(self) -> dict_values:
        """Get autodiscovery messages"""
        return self._autodiscovery_messages.values()



"""Helper dir for DS2482."""
from boneio.helper.onewire.ds2482 import DS2482, DS2482_ADDRESS
from boneio.helper.onewire.onewire import (
    OneWire,
    OneWireBus,
    OneWireAddress,
    reverse_dallas_id,
)
from boneio.helper.onewire.W1ThermSensor import AsyncBoneIOW1ThermSensor


__all__ = [
    "DS2482",
    "OneWire",
    "OneWireBus",
    "DS2482_ADDRESS",
    "AsyncBoneIOW1ThermSensor",
    "OneWireAddress",
    "reverse_dallas_id",
]



import os

from PIL import ImageFont

from boneio.const import FONTS


def make_font(name: str, size: int, local: bool = False):
    """Prepare ImageFont for Oled screen."""
    font_path = (
        name
        if not local
        else os.path.join(os.path.dirname(__file__), "..", FONTS, name)
    )
    return ImageFont.truetype(font_path, size)



"""Filter class to adjust sensor values."""
from __future__ import annotations


FILTERS = {
    "offset": lambda x, y: x + y,
    "round": lambda x, y: round(x, y),
    "multiply": lambda x, y: x * y,
    "filter_out": lambda x, y: None if x == y else x,
    "filter_out_greater": lambda x, y: None if x > y else x,
    "filter_out_lower": lambda x, y: None if x < y else x,
}


class Filter:
    _filters = []

    def _apply_filters(self, value: float) -> float | None:
        for filter in self._filters:
            for k, v in filter.items():
                if k not in FILTERS:
                    break
                value = FILTERS[k](value, v)
                if value is None:
                    return None
        return value



import fnmatch
import logging
import os
import re
from collections import OrderedDict
from typing import Any, Tuple

from cerberus import TypeDefinition, Validator
from yaml import MarkedYAMLError, SafeLoader, YAMLError, load

from boneio.const import COVER, ID, OUTPUT
from boneio.helper.exceptions import ConfigurationException
from boneio.helper.timeperiod import TimePeriod

schema_file = os.path.join(os.path.dirname(__file__), "../schema/schema.yaml")
_LOGGER = logging.getLogger(__name__)

SECRET_YAML = "secrets.yaml"
_SECRET_VALUES = {}


class BoneIOLoader(SafeLoader):
    """Loader which support for include in yaml files."""

    def __init__(self, stream):

        self._root = os.path.split(stream.name)[0]

        super(BoneIOLoader, self).__init__(stream)

    def include(self, node):

        filename = os.path.join(self._root, self.construct_scalar(node))

        with open(filename, "r") as f:
            return load(f, BoneIOLoader)

    def _rel_path(self, *args):
        return os.path.join(self._root, *args)

    def construct_secret(self, node):
        secrets = load_yaml_file(self._rel_path(SECRET_YAML))
        if node.value not in secrets:
            raise MarkedYAMLError(f"Secret '{node.value}' not defined", node.start_mark)
        val = secrets[node.value]
        _SECRET_VALUES[str(val)] = node.value
        return val

    def represent_stringify(self, value):
        return self.represent_scalar(tag="tag:yaml.org,2002:str", value=str(value))

    def construct_include_dir_list(self, node):
        files = filter_yaml_files(_find_files(self._rel_path(node.value), "*.yaml"))
        return [load_yaml_file(f) for f in files]

    def construct_include_dir_merge_list(self, node):
        files = filter_yaml_files(_find_files(self._rel_path(node.value), "*.yaml"))
        merged_list = []
        for fname in files:
            loaded_yaml = load_yaml_file(fname)
            if isinstance(loaded_yaml, list):
                merged_list.extend(loaded_yaml)
        return merged_list

    def construct_include_dir_named(self, node):
        files = filter_yaml_files(_find_files(self._rel_path(node.value), "*.yaml"))
        mapping = OrderedDict()
        for fname in files:
            filename = os.path.splitext(os.path.basename(fname))[0]
            mapping[filename] = load_yaml_file(fname)
        return mapping

    def construct_include_dir_merge_named(self, node):
        files = filter_yaml_files(_find_files(self._rel_path(node.value), "*.yaml"))
        mapping = OrderedDict()
        for fname in files:
            loaded_yaml = load_yaml_file(fname)
            if isinstance(loaded_yaml, dict):
                mapping.update(loaded_yaml)
        return mapping


BoneIOLoader.add_constructor("!include", BoneIOLoader.include)
BoneIOLoader.add_constructor("!secret", BoneIOLoader.construct_secret)
BoneIOLoader.add_constructor(
    "!include_dir_list", BoneIOLoader.construct_include_dir_list
)
BoneIOLoader.add_constructor(
    "!include_dir_merge_list", BoneIOLoader.construct_include_dir_merge_list
)
BoneIOLoader.add_constructor(
    "!include_dir_named", BoneIOLoader.construct_include_dir_named
)
BoneIOLoader.add_constructor(
    "!include_dir_merge_named", BoneIOLoader.construct_include_dir_merge_named
)


def filter_yaml_files(files):
    return [
        f
        for f in files
        if (
            os.path.splitext(f)[1] in (".yaml", ".yml")
            and os.path.basename(f) not in ("secrets.yaml", "secrets.yml")
            and not os.path.basename(f).startswith(".")
        )
    ]


def _is_file_valid(name):
    """Decide if a file is valid."""
    return not name.startswith(".")


def _find_files(directory, pattern):
    """Recursively load files in a directory."""
    for root, dirs, files in os.walk(directory, topdown=True):
        dirs[:] = [d for d in dirs if _is_file_valid(d)]
        for basename in files:
            if _is_file_valid(basename) and fnmatch.fnmatch(basename, pattern):
                filename = os.path.join(root, basename)
                yield filename


def load_yaml_file(filename: str) -> Any:
    with open(filename, "r") as stream:
        try:
            return load(stream, Loader=BoneIOLoader) or OrderedDict()
        except YAMLError as exception:
            raise exception


def load_config_from_string(config_yaml: str):
    schema = load_yaml_file(schema_file)
    v = CustomValidator(schema, purge_unknown=True)
    v.validate(config_yaml)
    doc = v.normalized(v.document, always_return_document=True)
    # validated = v.validated(document=doc, normalize=True, always_return_document=True)
    # doc = v.normalized(validated, always_return_document=True)
    return doc


def load_config_from_file(config_file: str):
    try:
        config_yaml = load_yaml_file(config_file)
    except FileNotFoundError as err:
        raise ConfigurationException(err)
    if not config_yaml:
        _LOGGER.warn("Missing yaml file. %s", config_file)
        return None
    return load_config_from_string(config_yaml)


def one_of(*values, **kwargs):
    """Validate that the config option is one of the given values.
    :param values: The valid values for this type
    """
    options = ", ".join(f"'{x}'" for x in values)

    def validator(value):
        if value not in values:
            import difflib

            options_ = [str(x) for x in values]
            option = str(value)
            matches = difflib.get_close_matches(option, options_)
            if matches:
                matches_str = ", ".join(f"'{x}'" for x in matches)
                raise ConfigurationException(
                    f"Unknown value '{value}', did you mean {matches_str}?"
                )
            raise ConfigurationException(
                f"Unknown value '{value}', valid options are {options}."
            )
        return value

    return validator


timeperiod_type = TypeDefinition("timeperiod", (TimePeriod,), ())


class CustomValidator(Validator):
    """Custom validator of cerberus"""

    types_mapping = Validator.types_mapping.copy()
    types_mapping["timeperiod"] = timeperiod_type

    def _lookup_field(self, path: str) -> Tuple:
        """
        Implement relative paths with dot (.) notation, following Python
        guidelines: https://www.python.org/dev/peps/pep-0328/#guido-s-decision
        - A single leading dot indicates a relative import
        starting with the current package.
        - Two or more leading dots give a relative import to the parent(s)
        of the current package, one level per dot after the first
        Return: Tuple(dependency_name: str, dependency_value: Any)
        """
        # Python relative imports use a single leading dot
        # for the current level, however no dot in Cerberus
        # does the same thing, thus we need to check 2 or more dots
        if path.startswith(".."):
            parts = path.split(".")
            dot_count = path.count(".")
            context = self.root_document

            for key in self.document_path[:dot_count]:
                context = context[key]

            context = context.get(parts[-1])

            return parts[-1], context

        else:
            return super()._lookup_field(path)

    def _check_with_output_id_uniqueness(self, field, value):
        """Check if outputs ids are unique."""
        all_ids = [x[ID] for x in self.document[OUTPUT]]
        if len(all_ids) != len(set(all_ids)):
            self._error(field, "Output IDs are not unique.")

    def _normalize_coerce_to_bool(self, value):
        return True

    def _normalize_coerce_lower(self, value):
        return str(value).lower()

    def _normalize_coerce_upper(self, value):
        return str(value).upper()

    def _normalize_coerce_actions_output(self, value):
        return str(value).lower()

    def _normalize_coerce_str(self, value):
        return str(value)

    def _normalize_coerce_check_actions(self, value):
        _path = self.document_path
        parent = self.root_document[_path[0]][_path[1]]
        keys = value.keys()
        _schema = self.schema["actions"]
        out = {}
        for key in keys:
            _deps = _schema["schema"][key]["dependencies"].items()
            d = next(iter(_deps))
            _parent_key_value = parent.get(d[0], "switch")
            for _v in d[1]:
                if _v == _parent_key_value:
                    out[key] = value[key]
        return out

    def _normalize_coerce_positive_time_period(self, value) -> TimePeriod:
        """Validate and transform time period with time unit and integer value."""

        if isinstance(value, int):
            raise ConfigurationException(
                f"Don't know what '{value}' means as it has no time *unit*! Did you mean '{value}s'?"
            )
        if isinstance(value, TimePeriod):
            value = str(value)
        if not isinstance(value, str):
            raise ConfigurationException("Expected string for time period with unit.")

        unit_to_kwarg = {
            "us": "microseconds",
            "microseconds": "microseconds",
            "ms": "milliseconds",
            "milliseconds": "milliseconds",
            "s": "seconds",
            "sec": "seconds",
            "secs": "seconds",
            "seconds": "seconds",
            "min": "minutes",
            "minutes": "minutes",
            "h": "hours",
            "hours": "hours",
            "d": "days",
            "days": "days",
        }

        match = re.match(r"^([-+]?[0-9]*\.?[0-9]*)\s*(\w*)$", value)
        if match is None:
            raise ConfigurationException(f"Expected time period with unit, got {value}")
        kwarg = unit_to_kwarg[one_of(*unit_to_kwarg)(match.group(2))]

        return TimePeriod(**{kwarg: float(match.group(1))})

    def _normalize_coerce_check_action_def(self, value):
        parent = self.root_document[self.document_path[0]][self.document_path[1]]
        _schema = self.schema["actions"]
        keys = value.keys()
        out = {}
        for key in keys:
            _deps = _schema["schema"][key]["dependencies"].items()
            d = next(iter(_deps))
            _parent_key_value = parent[d[0]]
            for _v in d[1]:
                if _v == _parent_key_value:
                    out[key] = value[key]
        return out

    def _normalize_default_setter_toggle_cover(self, document):
        def get_parent():
            out = self.root_document
            for _p in self.document_path:
                out = out[_p]
            return out

        parent = get_parent()
        if parent["action"] == COVER:
            return "toggle"



"""ADC GPIO BBB sensor."""
from datetime import datetime
import logging

from boneio.const import SENSOR
from boneio.helper import BasicMqtt, AsyncUpdater
from boneio.helper.filter import Filter

try:
    import Adafruit_BBIO.ADC as ADC
except ModuleNotFoundError:

    class ADC:
        def __init__(self):
            pass

    pass

_LOGGER = logging.getLogger(__name__)


def initialize_adc():
    ADC.setup()


class GpioADCSensor(BasicMqtt, AsyncUpdater, Filter):
    """Represent Gpio ADC sensor."""

    def __init__(self, pin: str, **kwargs) -> None:
        """Setup GPIO ADC Sensor"""
        super().__init__(topic_type=SENSOR, **kwargs)
        self._pin = pin
        self._state = None
        AsyncUpdater.__init__(self, **kwargs)
        _LOGGER.debug("Configured sensor pin %s", self._pin)

    @property
    def state(self) -> float:
        """Give rounded value of temperature."""
        return self._state

    def update(self, time: datetime) -> None:
        """Fetch temperature periodically and send to MQTT."""
        _state = self._apply_filters(value=ADC.read(self._pin))
        if not _state:
            return
        self._state = _state
        self._send_message(
            topic=self._send_topic,
            payload=self.state,
        )



"""Manage BoneIO onboard temp sensors."""

import asyncio
from datetime import datetime

from boneio.const import SENSOR, STATE, TEMPERATURE
from boneio.helper import BasicMqtt, AsyncUpdater
from boneio.helper.exceptions import I2CError
from boneio.helper.filter import Filter


class TempSensor(BasicMqtt, AsyncUpdater, Filter):
    """Represent Temp sensor in BoneIO."""

    SensorClass = None
    DefaultName = TEMPERATURE

    def __init__(
        self,
        i2c,
        address: str,
        id: str = DefaultName,
        filters: list = ["round(x, 2)"],
        **kwargs
    ):
        """Initialize Temp class."""
        super().__init__(id=id, topic_type=SENSOR, **kwargs)
        self._loop = asyncio.get_event_loop()
        self._filters = filters
        try:
            self._pct = self.SensorClass(i2c_bus=i2c, address=address)
            self._state: float | None = None
        except ValueError as err:
            raise I2CError(err)
        AsyncUpdater.__init__(self, **kwargs)

    @property
    def state(self) -> float:
        """Give rounded value of temperature."""
        return self._state

    def update(self, time: datetime) -> None:
        """Fetch temperature periodically and send to MQTT."""
        _temp = self._apply_filters(value=self._pct.temperature)
        if not _temp:
            return
        self._state = _temp
        self._send_message(
            topic=self._send_topic,
            payload={STATE: self._state},
        )



"""Class to help initialize classes which uses mqtt send."""
from typing import Callable, Union


class BasicMqtt:
    """Basic MQTT class."""

    def __init__(
        self,
        id: str,
        topic_prefix: str,
        name: str,
        send_message: Callable[[str, Union[str, dict]], None],
        topic_type: str,
        **kwargs,
    ):
        """Initialize module."""
        self._id = id.replace(" ", "")
        self._name = name
        self._send_message = send_message
        self._send_topic = f"{topic_prefix}/{topic_type}/{self.id}"

    @property
    def id(self) -> bool:
        """Id of the module."""
        return self._id

    @property
    def name(self):
        """Return name of the sensor."""
        return self._name
